"""
This file is part of LionsTooth.

LionsTooth is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

LionsTooth is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with LionsTooth.  If not, see <http://www.gnu.org/licenses/>.
"""

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtWidgets import QFileDialog, QInputDialog
from PyQt5.QtCore import QObject, QThreadPool, QRunnable, pyqtSignal, pyqtSlot

""" Image resolver / caching """
from classes.resolver import resolver

""" Diaspy """
import diaspy
from diaspy.errors import DiaspyError

""" Exceptions """
import requests
"""ReadTimeout requires minimal version of python3-requests 2.11 
https://pypi.python.org/pypi/requests/2.11.1 """
version = requests.__version__.split('.')
if int(version[0]) == 2 and int(version[1]) >= 11:
	from requests import ConnectionError, ReadTimeout
else:
	from requests import ConnectionError
	class ReadTimeout(Exception):
		pass

class stream(diaspy.streams.Stream):
	def __init__(self, connection, location=''):
		self.connection = connection
		self.location = location
		diaspy.streams.Stream.__init__(self, self.connection, location=self.location, fetch=False)

		self.layout = None
		self.post_widgets = {}

		""" TODO implement this
			If more() doesn't give more posts then block future calls.
			So it doesn't keep requesting when there are no more posts.
		"""
		self.no_more = False

	def set_layout(self, layout):
		self.layout = layout

class aspects(diaspy.streams.Aspects):
	def __init__(self, connection):
		self.connection = connection
		diaspy.streams.Aspects.__init__(self, self.connection, fetch=False)

		self.layout = None
		self.post_widgets = {}
		self.no_more = False

	def set_layout(self, layout):
		self.layout = layout

class handler(QObject):
	"""
	Threads and Object storage.
	"""
	message = pyqtSignal(tuple) # title, message
	connected = pyqtSignal(bool)
	auth_status = pyqtSignal(bool)
	userdata_ready = pyqtSignal()
	tag_followings_ready = pyqtSignal()
	new_stream = pyqtSignal(str) #  str: name
	new_notifications_obj = pyqtSignal()
	new_people_obj = pyqtSignal(str) # str: guid
	def __init__(self):
		QObject.__init__(self)

		self.connection = None
		self.search = None

		# Cache people data by guid { guid: peope obj }
		self.people = {} # will contain people data, inc streams TODO move to people page.
		self.people_handle_to_guid = {} # key: handle, value: guid

		self.notifications = None

		# Thread pools
		self.connection_pool = QThreadPool()
		self.connection_pool.setMaxThreadCount(2)

		self.stream_pool = QThreadPool()
		self.stream_pool.setMaxThreadCount(3)

		self.notification_pool = QThreadPool()
		self.notification_pool.setMaxThreadCount(2)

		self.people_pool = QThreadPool()
		self.people_pool.setMaxThreadCount(2)

		self.people_stream_pool = QThreadPool()
		self.people_stream_pool.setMaxThreadCount(1)

		self.tag_followings = None
		self.userdata = None
		self.contacts = None # expects diaspy.people.Contacts(connection)
		self.res = None
		self.publisher = None

	def set_contacts(self):
		# Just init, so connection is there and we are ready to use it.
		self.contacts = diaspy.people.Contacts(self.connection, fetch=False)

	def add_people_obj(self, guid, people_obj):
		if people_obj and guid:
			self.people.update( { guid: people_obj } )
			if people_obj.data['handle']: # TODO if requested through guid, there is no diaspora_id but there should be in data
				self.people_handle_to_guid.update( { people_obj.data['handle'] : guid } )
			self.new_people_obj.emit(guid)

	def set_notifications_obj(self, notifications_obj):
		if notifications_obj:
			self.notifications = notifications_obj
			self.new_notifications_obj.emit()

	def set_stream(self, name, stream):
		if stream:
			self.stream[name] = stream
			self.new_stream.emit(name)

	def set_connection(self, connection):
		if connection:
			self.connection = connection
			self.search = diaspy.search.Search( connection )

	def if_connected(self):
		print("called diaspy_handler.if_connected()") # DEBUG
		if self.connection:
			self.res = resolver( self.connection.pod ) # set resolver for images
			self.publisher = diaspy.publisher.Publisher( self.connection )
			self.connected.emit(True)
		else:
			self.connected.emit(False)

	def connect(self, config):
		print("called diaspy_handler.connect()") # DEBUG
		worker = connection_worker('connect', config=config)
		worker.signals.new_result.connect(self.set_connection)
		worker.signals.finished.connect(self.if_connected)
		self.connection_pool.start(worker)	

	def disconnect(self):
		self.connection.logout()
		self.connection = None
		self.res = None

	def login(self):
		worker = connection_worker('login', connection=self.connection)
		worker.signals.status_change.connect( self.auth_status.emit )
		self.connection_pool.start(worker)

	def set_user_data(self, userdata):
		if userdata:
			#self.userdata = userdata# TODO
			self.userdata_ready.emit()

	def get_user_data(self):
		worker = connection_worker('get_user_data', connection=self.connection)
		worker.signals.new_result.connect( self.set_user_data )
		self.connection_pool.start(worker)

	def set_tag_followings(self, tag_followings):
		if tag_followings:
			self.tag_followings = tag_followings
			self.tag_followings_ready.emit()

	def get_tag_followings(self):
		worker = connection_worker('get_tag_followings', connection=self.connection)
		worker.signals.new_result.connect( self.set_tag_followings )
		self.connection_pool.start(worker)

	def init_notifications(self):
		self.notifications = notifications(self.connection.__repr__())

	def is_connected(self):
		if self.connection:
			return True
		return False

""" Post worker """
class post_signals(QObject):
	finished = pyqtSignal()
	exception = pyqtSignal(str)
	new_post = pyqtSignal(object) # diaspy.models.Post
	new_comment = pyqtSignal(object) # diaspy.models.Comment
	success = pyqtSignal()

	def disconnect(self):
		self.finished.disconnect()
		self.exception.disconnect()
		self.new_post.disconnect()

class post_worker(QRunnable):
	def __init__(self, action, connection=None, post=None, guid=None, id=None, interaction=None, vote_answer_id=None, comment_markdown=None, delete_comment_id=None):
		super(post_worker, self).__init__()
		self.action = action.lower()
		self._interaction = interaction
		self.vote_answer_id = vote_answer_id
		self.delete_comment_id = delete_comment_id
		self.comment_markdown = comment_markdown
		self.__connection = connection
		self.post = post
		self.guid = guid
		self.id = id
		self.signals = post_signals()

	def get(self):
		try:
			if self.guid:
				post = diaspy.models.Post(self.__connection, guid=self.guid)
			elif self.id:
				post = diaspy.models.Post(self.__connection, id=self.id)
			else:
				pass # raise error
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error: {}".format(err))
		except diaspy.errors.PostError as err:
			print("Could not get post guid:{}, id:{}, err:{}".format(self.guid, self.id, err)) #DEBUG
			self.signals.exception.emit("Could not get post guid:{}, id:{}, err:{}".format(self.guid, self.id, err))
		else:
			self.signals.new_post.emit( post )
		finally:
			self.signals.finished.emit()

	def interaction(self):
		""" needs to have self._interaction  and self.post set
		self._interaction should be a str containing one of the 
		following:

			like
			unlike
			subscribe
			unsubscribe
			reshare
			hide
			delete
			mute
			report
			vote_poll

		self.post should be a diaspy.models.Post() object
		"""	

		try:
			if self._interaction == 'like': self.post.like()
			elif self._interaction == 'update': self.post.update() # also updates comments
			elif self._interaction == 'vote_poll': self.post.vote_poll( self.vote_answer_id ) # requires answer_id
			elif self._interaction == 'unlike': self.post.delete_like()
			elif self._interaction == 'subscribe': self.post.subscribe()
			elif self._interaction == 'unsubscribe': self.post.unsubscribe()
			elif self._interaction == 'reshare': self.post.reshare()
			elif self._interaction == 'hide': self.post.hide()
			elif self._interaction == 'delete': self.post.delete()
			elif self._interaction == 'mute': self.post.mute()
			elif self._interaction == 'report': self.post.report()
			elif self._interaction == 'delete_comment': self.post.comments.delete_comment(self.delete_comment_id)
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error, could not {} post: {}".format(self._interaction, err))
		except diaspy.errors.PostError as err: # probably post is deleted.
			self.signals.exception.emit("PostError: could not {} post: {}".format(self._interaction, err))
		else:
			self.signals.success.emit()
		finally:
			self.signals.finished.emit()

	def comment(self):
		try:
			c = self.post.comments.comment(text=self.comment_markdown)
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error, could comment on post: {}".format(err))
		else:
			self.signals.new_comment.emit( c )
		finally:
			self.signals.finished.emit()

	def run(self):
		print("called post_worker.run() {} {}".format(self.action, self._interaction))#DEBUG
		actions = {
			'get'			: self.get,
			'interaction'	: self.interaction,
			'comment'		: self.comment
		}[self.action]()

""" Stream worker """
class stream_signals(QObject):
	finished = pyqtSignal()
	exception = pyqtSignal(str)
	new_stream = pyqtSignal(object) # object: diaspy.stream.Generic()
	new_post = pyqtSignal(object) # object: diaspy.models.Post()
	new_data = pyqtSignal(str) # str: name

class stream_worker(QRunnable):
	provider_display_name = 'LionsTooth (development)'
	def __init__(self, action, connection=None, name='', location='', stream=None, post_data=None, aspect_ids_filter=None):
		super(stream_worker, self).__init__()
		self.action = action.lower()
		self.location = location
		self.name = name
		self.aspect_ids_filter = aspect_ids_filter

		""" self.post_data = {
			'text'			: str(markdown), 
			'aspect_ids'	: [ids], 
			'photos			: None, 
			'photo'			: '',
			'poll_question'	: 'poll_question',
			'poll_answers'	: ['answer0', 'answer1', ..]
		} """
		self.post_data = post_data
		self.__connection = connection
		self.stream = stream
		self.signals = stream_signals()

	def new_aspects(self):
		try:
			print("stream_worker->new_aspects()")#DEBUG
			new_stream = aspects(self.__connection)
			#new_stream.fill()
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error: {}".format(err))
		except DiaspyError as err:
			self.signals.exception.emit("Diaspy error: {}".format(err))
		else:
			self.signals.new_stream.emit(new_stream)
		finally:
			self.signals.finished.emit()

	def aspects_filter(self):# used on Aspects stream
		try:
			self.stream.filter( self.aspect_ids_filter )
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error: {}".format(err))
		finally:
			self.signals.finished.emit()

	def new(self):
		try:
			new_stream = stream(self.__connection, location=self.location)
			new_stream.fill()
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error: {}".format(err))
		except DiaspyError as err:
			self.signals.exception.emit("Diaspy error: {}".format(err))
		else:
			self.signals.new_stream.emit(new_stream)
		finally:
			self.signals.finished.emit()

	def fill(self):
		try:
			self.stream.fill()
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error: {}".format(err))
		else:
			self.signals.new_data.emit(self.name)
		finally:
			self.signals.finished.emit()

	def more(self):
		try:
			self.stream.more()
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error: {}".format(err))
		else:
			self.signals.new_data.emit(self.name)
		finally:
			self.signals.finished.emit()

	def update(self):
		try:
			self.stream.update()
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error: {}".format(err))
		else:
			self.signals.new_data.emit(self.name)
		finally:
			self.signals.finished.emit()

	def post(self):
		try:
			self.post = self.stream.post(
				text=self.post_data['text'], 
				aspect_ids=self.post_data['aspect_ids'], 
				photos=self.post_data['photos'], 
				photo=self.post_data['photo'],
				poll_question=self.post_data['poll_question'],
				poll_answers=self.post_data['poll_answers'],
				provider_display_name=self.provider_display_name )
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error: {}".format(err))
		else:
			self.signals.new_post.emit( self.post )
		finally:
			self.signals.finished.emit()

	def run(self):
		print("called stream_worker.run() {}".format(self.action))#DEBUG
		actions = {
			'new'			: self.new,
			'post'			: self.post,
			'fill'			: self.fill,
			'more'			: self.more,
			'update'		: self.update,
			'new_aspects'	: self.new_aspects,
			'aspects_filter': self.aspects_filter
		}[self.action]()

""" people worker """
class people_signals(QObject):
	finished = pyqtSignal()
	exception = pyqtSignal(str)
	new_obj = pyqtSignal(str, object) # str: guid, object: diaspy.people.User
	new_data = pyqtSignal(str) # str: guid

class people_worker(QRunnable):
	def __init__(self, action, guid=None, diaspora_id=None, connection=None, people_obj=None):
		super(people_worker, self).__init__()
		self.action = action
		self.guid = guid
		self.diaspora_id = diaspora_id
		self.people_obj = people_obj
		self.__connection = connection
		self.signals = people_signals()

	def add_people_obj(self):
		try:
			people_obj = diaspy.people.User(self.__connection, guid=self.guid, handle=self.diaspora_id, fetch='data')
			people_obj.getPhotos() # TODO?
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error: {}".format(err))
		else:
			self.signals.new_obj.emit(people_obj.data['guid'], people_obj)
		finally:
			self.signals.finished.emit()

	def get_photos(self):
		try:
			self.people_obj.getPhotos()
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error: {}".format(err))
		finally:
			self.signals.finished.emit()

	def get_userdata(self):
		try:
			self.people_obj.fetchguid(fetch_stream=False)
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error: {}".format(err))
		else:
			self.signals.new_data.emit(self.guid)
		finally:
			self.signals.finished.emit()

	def fetch_stream(self):
		try:
			self.people_obj._fetchstream()
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error: {}".format(err))
		else:
			self.signals.new_data.emit(self.guid)
		finally:
			self.signals.finished.emit()

	def run(self):
		print("called people_worker.run() {}".format(self.action))#DEBUG
		actions = {
			'add_people_obj'	: self.add_people_obj,
			'get_userdata'		: self.get_userdata,
			'fetch_stream'		: self.fetch_stream,
			'get_photos'		: self.get_photos
		}[self.action]()

""" Connection worker """
class connection_signals(QObject):
	finished = pyqtSignal()
	exception = pyqtSignal(str)
	new_result = pyqtSignal(object)
	status_change = pyqtSignal(bool)

class connection_worker(QRunnable):
	def __init__(self, action, connection=None, config=None):
		super(connection_worker, self).__init__()
		self.action = action.lower() # make sure it's always lowercase
		self.__connection = connection
		self.__config = config
		self.signals = connection_signals()

	def connect(self):
		print("called connection_worker.connect()")#DEBUG
		try:
			pod = "https://{}".format(self.__config['pod'])
			connection = diaspy.connection.Connection(pod=pod, username=self.__config['user'], password=self.__config['pass'])
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error: {}".format(err))
		else:
			self.signals.new_result.emit(connection)
		finally:
			self.signals.finished.emit()

	def login(self):
		try:
			self.__connection.login()
		except diaspy.errors.LoginError as err:
			self.signals.exception.emit("login() failed: {}".format(err))
			self.signals.status_change.emit(False)
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error: {}".format(err))
			self.signals.status_change.emit(False)
		else:
			self.signals.status_change.emit(True)
		finally:
			self.signals.finished.emit()

	def get_user_data(self):
		try:
			userdata = self.__connection.getUserData()
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error on getUserData(): {}".format(err))
		else:
			self.signals.new_result.emit(userdata)
		finally:
			self.signals.finished.emit()

	def get_tag_followings(self):
		try:
			#following_tags = self.__connection.getTagFollowings()
			following_tags = diaspy.tags.Tags(self.__connection)
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error on getTagFollowings(): {}".format(err))
		else:
			self.signals.new_result.emit(following_tags)
		finally:
			self.signals.finished.emit()

	def run(self):
		print("called connection_worker.run() {}".format(self.action)) # DEBUG
		actions = {
			'connect'				: self.connect,
			'login'					: self.login,
			'get_user_data'			: self.get_user_data,
			'get_tag_followings'	: self.get_tag_followings
		}[self.action]()

""" notification worker """
class notifications_signals(QObject):
	finished = pyqtSignal()
	exception = pyqtSignal(str)
	new_obj = pyqtSignal(object) # object: diaspy.notifications.Notifications
	new_data = pyqtSignal()

class notifications_worker(QRunnable):
	def __init__(self, action, connection=None, notifications_obj=None):
		super(notifications_worker, self).__init__()
		self.action = action
		self.__connection = connection
		self.signals = notifications_signals()
		self.notifications = notifications_obj

	def new(self):
		try:
			new_notifications = diaspy.notifications.Notifications(self.__connection)
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error on getTagFollowings(): {}".format(err))
		else:
			self.signals.new_obj.emit(new_notifications)
		finally:
			self.signals.finished.emit()

	def more(self):
		try:
			self.notifications.more()
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error on getTagFollowings(): {}".format(err))
		else:
			self.signals.new_data.emit()
		finally:
			self.signals.finished.emit()

	def update(self):
		try:
			self.notifications.update()
		except (ConnectionError, ReadTimeout) as err:
			self.signals.exception.emit("Connection error on getTagFollowings(): {}".format(err))
		else:
			self.signals.new_data.emit()
		finally:
			self.signals.finished.emit()

	def run(self):
		print("called notifications_worker.run() {}".format(self.action))#DEBUG
		actions = {
			'new'		: self.new,
			'more'		: self.more,
			'update'	: self.update
		}[self.action]()
