"""
This file is part of LionsTooth.

LionsTooth is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

LionsTooth is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with LionsTooth.  If not, see <http://www.gnu.org/licenses/>.
"""

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtWidgets import QFileDialog, QInputDialog, QLabel
from PyQt5.QtCore import pyqtSignal, pyqtSlot, QRect

""" UI """
from design.ui_main_window import Ui_MainWindow

""" Widgets """
from widgets.header import header
from widgets.post_widget import post_widget
from widgets.podlist import podlist
from widgets.side_menu import side_menu
from widgets.msg import message
from widgets.notifications import notifications

""" Settings Widgets """
from widgets.settings.cache import cache as cache_widget

""" Pages """
import pages

""" Reading and writing config """
import configparser

""" For getting image dimensions """
#from PIL import Image

""" Searching and replacing image tags / attr """
import re

""" Diaspy handler """
from classes.handler import handler, stream_worker, notifications_worker, people_worker, post_worker

""" Exceptions """
from requests import ConnectionError

""" Debugging """ # http://code.activestate.com/recipes/579105-how-a-python-function-can-find-the-name-of-its-cal/
import sys
getframe_expr = 'sys._getframe({}).f_code.co_name'

""" Parsing HTML from notifications """
from bs4 import BeautifulSoup

from classes.resolver import img_worker
from classes.workarounds import workarounds

class Window(QtWidgets.QMainWindow):
	resized = pyqtSignal()

	def  __init__(self, parent=None):
		super(Window, self).__init__(parent=parent)
		self.ui = Ui_MainWindow()
		self.ui.setupUi(self)

		self.workarounds = workarounds()

		""" overlay widget test """
		self.side_menu = side_menu(self)
		self.ui.verticalLayout.setContentsMargins(self.side_menu.min_width, 0, 0, 0) # Set left margin of horizontalLayout
		self.side_menu.show()

		""" display messages to user """
		self.msg = message(self)
		self.msg.update_size( size=self.size(), margins=self.ui.verticalLayout.contentsMargins() )

		self.pl = None # podlist
		self.resize_running = False

		""" Config """
		self._config_file = 'config.ini'
		self._config = configparser.ConfigParser()
		self._config.read(self._config_file)

		# Connection
		if 'connection' not in self._config:
			self._config['connection'] = {}
			self._config['connection']['verified'] = 'False'
		else:
			if 'pod' in self._config['connection']: self.ui.line_pod.setText( self._config['connection']['pod'] )
			if 'user' in self._config['connection']: self.ui.line_username.setText( self._config['connection']['user'] )
			if 'pass' in self._config['connection']: self.ui.line_password.setText( self._config['connection']['pass'] )
			if 'verified' not in self._config['connection']: self._config['connection']['verified'] = 'False'

		# General
		size = self.size()
		if 'general' not in self._config:
			self._config['general'] = {}
			self._config['general']['width'] = str(size.width())
			self._config['general']['height'] = str(size.height())
		else:
			width, height = size.width(), size.height()
			if 'width' in self._config['general']:
				if int(self._config['general']['height']) in range(100,1200):
					width = int(self._config['general']['width'])
			if 'height' in self._config['general']:
				if int(self._config['general']['height']) in range(100,1200):
					height = int(self._config['general']['height'])
			self.resize(width, height)

		# Cache
		if 'cache' not in self._config:
			self._config['cache'] = {
				'cache_dir'			: './cache',
				'clear_cache_days'	: 7,
				'clear_all_cache'	: False,
				'clear_on_start'	: True
			}

		""" Diaspy handler with QThreadPool's """
		self.handle = handler()

		""" Init pages """
		""" Pages in self.stream_pages are stream pages, thus those pages that 
		are a stream widget. """
		self.stream_pages = {}
		self.stream_pages.update( { 'people': pages.people.people( thread_pool=self.handle.people_pool ) } )
		self.stream_pages.update( { 'stream': pages.stream.stream( thread_pool=self.handle.stream_pool ) } )
		self.stream_pages.update( { 'aspects': pages.aspects.aspects( thread_pool=self.handle.stream_pool ) } )
		self.stream_pages.update( { 'mentions': pages.mentions.mentions( thread_pool=self.handle.stream_pool ) } )
		self.stream_pages.update( { 'activity': pages.activity.activity( thread_pool=self.handle.stream_pool ) } )
		self.stream_pages.update( { 'public': pages.public.public( thread_pool=self.handle.stream_pool ) } )
		self.stream_pages.update( { 'followed_tags': pages.followed_tags.followed_tags( thread_pool=self.handle.stream_pool ) } )
		self.stream_pages.update( { 'single_post': pages.single_post.single_post( thread_pool=self.handle.stream_pool ) } )

		for page_name in self.stream_pages:
			self.ui.stackedWidget.addWidget( self.stream_pages[page_name] )
			self.stream_pages[page_name].anchor_clicked.connect( self.on_hyperlink ) # hyperlinks

		""" Normal pages """
		self.pages = {}
		self.pages.update( { 'contacts' : pages.contacts.contacts() } )
		self.pages.update( { 'people_results' : pages.people_results.people_results() } )
		self.pages.update( { 'tag_results' : pages.tag_results.tag_results() } )

		for name, page in self.pages.items():
			self.ui.stackedWidget.addWidget( page )

		# Place settings here, 'cause it's already added
		self.pages.update( { 'settings' : self.ui.stackedWidget.widget(0) } )

		""" Settings """
		# Cache
		self.ui.cache_widget = cache_widget( config=self._config['cache'] )
		self.ui.cache_widget.config_updated.connect( self.config_updated )
		self.ui.settings_layout.insertWidget(1, self.ui.cache_widget )

		# Behaviour
		# ..

		""" Set header on stream pages """
		self.ui.header = header()
		self.ui.verticalLayout.insertWidget(0, self.ui.header)

		""" notifications """
		self.notifications = notifications(self)
		self.ui.header.ui.btn_notify.updateGeometry()
		self.notifications.update_size( size=self.size(), margins=self.ui.verticalLayout.contentsMargins() )

		""" Hooks """
		# Side menu buttons
		self.side_menu.selected.connect( self.change_page ) # button name and page name must be equal

		# Side menu adjust height if main is resized
		self.resized.connect( lambda: self.side_menu.adjust_height(self.height()) )

		# Settings page buttons
		self.ui.btn_save.clicked.connect(self.save_connection_settings)
		self.ui.btn_open_podlist.clicked.connect(self.handle_podlist_state)
		self.ui.btn_connect.clicked.connect(self.connect)

		self.stream_page_action = { # button actions
			'followed_tags' : self.on_tags_button,
			'aspects'		: self.on_aspect_button,
			'activity'		: self.on_activity_button,
			'mentions'		: self.on_mentions_button,
			'public'		: self.on_public_button,
		}

		self.page_action = { # button acions
			'profile' : self.on_people_event,
		}

		""" Resize posts HTML/photo_browser images when Window has resized
		Also does post_widget.html_content widget resize """
		self.resized.connect(lambda: self.resize_event())

		""" do on start """
		self.side_menu.ui.btn_side_menu.click() # colapse side menu on start
		#self.side_menu.stream_selector.ui.btn_stream.click() # set start page to stream

		if self._config['connection']['verified'] == 'True':
			self.ui.btn_connect.click() # connect on start, needs some of checks TODO
		else:
			self.msg.print_msg("Welcome to LionsTooth", """<h3>Please be aware this project still is in a very early stage, lacks features and will contain various bugs!</h3><br/><br/>Pod settings not verified yet. If this is your first time running LionsTooth you'd have to set a pod, username and password in the settings tab. Press connect after, if you want to connect. Have fun ;-)""" )

		""" TMP """
		self.thread_list = {
			'stream': {
				'more'			: [],
				'update'		: [],
			}
		}

	""" Config """
	def config_updated(self, key, value):
		for k in value:
			self._config[key][k] = str(value[k])
		with open(self._config_file, 'w') as configfile:
			self._config.write(configfile)

	""" Common function """
	def clear_layout(self, layout, leave_at_first=0, leave_at_end=0): # https://riverbankcomputing.com/pipermail/pyqt/2010-August/027450.html
		while layout.count()-leave_at_end > leave_at_first:
			item = layout.takeAt(leave_at_first)
			if not item:
				continue
			w = item.widget()
			if w:
				w.deleteLater()

	def change_page(self, name):
		if name in self.pages:
			self.ui.stackedWidget.setCurrentIndex( self.ui.stackedWidget.indexOf(self.pages[name]) )

		if name in self.page_action:
			self.page_action[name]()

	def change_stream_page(self, name):
		if name in self.stream_pages:
			self.ui.stackedWidget.setCurrentIndex( self.ui.stackedWidget.indexOf(self.stream_pages[name]) )

		if name in self.stream_page_action:
			self.stream_page_action[name]()

	def leave_page(self, name):
		# Stop animations if needed
		# Create self.stream_pages_leave_acion[name] and use it as switch
		pass

	def open_settings(self, state):
		if state:
			self.ui.stackedWidget.setCurrentIndex(0)
			self.side_menu.stream_selector.deselect()

	def new_stream(self, name, location='', aspects=False):
		if aspects:
			worker = stream_worker('new_aspects', connection=self.handle.connection, name=name)
		else:
			worker = stream_worker('new', connection=self.handle.connection, name=name, location=location)
		worker.signals.new_stream.connect( self.stream_pages[name].set_stream )
		worker.signals.finished.connect(self.stream_pages[name].stop_top_loading)
		worker.signals.finished.connect(lambda: print("finished new_stream thread {}".format(name)))

		if name == 'followed_tags': # TODO implement this in stream widget itself
			worker.signals.finished.connect( self.stream_pages[name].set_not_busy )

		self.stream_pages[name].new_stream.connect( self.stream_pages[name].insert_post_widgets )

		# TODO disconnect above signals when finished.

		if name == 'followed_tags':
			self.stream_pages['followed_tags'].is_busy = True

		self.handle.stream_pool.start(worker)
		self.stream_pages[name].start_top_loading()

	def message(self, tuple):
		self.msg.print_msg(tuple[0], tuple[1])

	def on_hyperlink(self, link):
		m = re.match('^(tag|diaspora_id|https?)://(.*?)$', link)
		m = self.workarounds.groups_to_substitution(m) # workaround for different versions of regex
		if m:
			if m[1] == 'tag':
				self.open_tag_page( m[2].lower() ) # for some reason tag name must be in lower case else the request will return None
				self.side_menu.stream_selector.deselect() # TODO should not be here.
			elif m[1] == 'diaspora_id':
				self.on_people_event( diaspora_id=m[2] )
				self.side_menu.stream_selector.deselect() # TODO should not be here.
		else:
			self.on_diaspora_link( link )
		print(link)#DEBUG

	""" single post """
	def on_single_post_event(self, guid=None, id=None, comment_guid=None):
		"""
		param guid: post guid to fetch
		type: str

		param id: post id to fetch
		type: int

		param comment_guid: comment_guid to set on top in stream view 
						(selected comment). This won't be used to fetch
						anything. But we pass this value to 
						self.stream_pages['single_post'].go_to_anchor(post_guid=guid, comment_id=comment_guid)
		type: int for id or str for guid
		"""
		if guid:
			worker = post_worker('get', connection=self.handle.connection, guid=guid)
			self.stream_pages['single_post'].go_to_anchor_que.append([guid, comment_guid])
		elif id:
			worker = post_worker('get', connection=self.handle.connection, id=id)
			self.stream_pages['single_post'].go_to_anchor_que.append([id, comment_guid])

		worker.signals.new_post.connect( self.stream_pages['single_post'].set_post )
		self.handle.stream_pool.start(worker)
		self.change_stream_page('single_post')

	""" notifications """
	def on_diaspora_link(self, link):
		# TODO Parse links like https://{pod_netloc}/posts/{post_guid} and change them to permalinks
		print("on_diaspora_link: {}".format(link))
		try:
			_type, uid = link.lstrip('/').split('/', 1)
			comment_guid = None
			test_comment_guid = uid.split('#', 1)
			if len(test_comment_guid) == 2:
				uid = test_comment_guid[0]
				comment_guid = test_comment_guid[1]
		except ValueError:
			self.msg.print_msg("Could not open link", "Link `{}` is invalid.".format(link))
		else:
			if _type == 'posts':
				self.on_single_post_event(id=uid, comment_guid=comment_guid)

			elif _type == 'people':
				self.on_people_event(guid=uid)

	def on_notifications_button(self):
		if self.handle:
			if self.notifications.isVisible():
				self.notifications.hide()
				return
			else:
				pos = self.pos()
				self.notifications.move( pos.x()+60, pos.y()+self.notifications.top_margin )
				self.notifications.show()

			if not self.handle.notifications:
				worker = notifications_worker('new', connection=self.handle.connection)
				worker.signals.new_obj.connect( self.handle.set_notifications_obj )
				self.handle.new_notifications_obj.connect( self.new_notifications )
				self.handle.notification_pool.start(worker)

				self.notifications.request_more.connect( self.get_more_notification )
				self.notifications.anchor_clicked.connect( self.on_diaspora_link )

				self.notifications.notification_toggled.connect( self.ui.header.notification_toggled )# update header count when notification is toggled
			else:
				worker = notifications_worker('update', notifications_obj=self.handle.notifications)
				worker.signals.new_data.connect( self.new_notifications )
				self.handle.notification_pool.start(worker)

	def get_more_notification(self):
		worker = notifications_worker('more', notifications_obj=self.handle.notifications)
		worker.signals.new_data.connect( self.new_notifications )
		self.handle.notification_pool.start(worker)

	def parse_notification_html(self, n):
		"""
		Diaspora returns a notification in HTML, we are going to use it
		but first lets remove some elements that we will set on the 
		people page's profile from diaspy data.
		"""
		soup = BeautifulSoup(n._data['note_html'], 'lxml')
		img = soup.find('img', {"class": "avatar"})
		avatar_url = None
		people = {}

		""" remove timeago """
		timeago = soup.find('time', {"class": "timeago"})
		timeago_div = timeago.find_parent('div')
		timeago_div.decompose()

		""" remove unread toggle """
		unread_toggle = soup.find('div', {"class": "unread-toggle"})
		unread_toggle.decompose()

		""" avatar """
		if img.has_attr('src'):
			avatar_url = img['src']

		""" remove avatar """
		avatar = soup.find('div', {"class": ["media-object", "pull-left"]})
		avatar.decompose()

		""" people """
		links = soup.findAll('a', {"class": "hovercardable"})
		for link in links:
			if link.has_attr('data-hovercard'):
				guid = link['href'].rsplit('/', 1)[1]
				people.update( { guid: link.text } )
		return [avatar_url, people, str(soup)]

	def new_notifications(self):
		notifications = self.handle.notifications._notifications
		for i in range(len(notifications)): # handler class
			if notifications[i].id not in self.notifications.notifications:
				avatar_url, people, html = self.parse_notification_html(notifications[i]) # parse html

				self.notifications.insert(notifications[i], i) # this is the widget
				self.notifications.notifications[notifications[i].id].set_text( html )

				worker = img_worker(self.handle.res, url=avatar_url)
				worker.signals.finished.connect(self.notifications.notifications[notifications[i].id].update_avatar)
				self.handle.res.thread_pool.start(worker)

		self.ui.header.set_notification_count( self.handle.notifications._data['unread_count'] ) # set unread notification count in header

	""" activity """
	def on_activity_button(self):
		if self.handle:
			if not self.stream_pages['activity'].stream:
				self.new_stream('activity', location='activity.json')

	""" mentions """
	def on_mentions_button(self):
		if self.handle:
			if not self.stream_pages['mentions'].stream:
				self.new_stream('mentions',	location='mentions.json')

	""" public """
	def on_public_button(self):
		if self.handle:
			if not self.stream_pages['public'].stream:
				self.new_stream('public', location='public.json')

	""" Get tag following
	This is a list of followed tags
	Will be called when Tags button in side menu is pressed """
	def on_got_tag_followings(self):
		if self.handle.tag_followings:
			self.side_menu.stream_selector.new_tag_followings( self.handle.tag_followings )
			self.handle.tag_followings_ready.disconnect()
			self.side_menu.stream_selector.request_follow_tag.connect( self.on_new_tag ) # new tag request signal
			self.side_menu.stream_selector.open_tags()

	def del_tag_from_page(self, tag_id, tag_name): # TODO use the id instead
		self.side_menu.stream_selector.del_tag(tag_id)
		self.stream_pages['followed_tags'].header.tag_unfollowed()

	def on_new_tag(self, name, from_page=False):
		try:
			self.handle.tag_followings.follow(name)
		except:
			self.msg.print_msg("Failed", "Failed to add '#{}'. Please wait a few seconds before adding another tag.".format(name) )
		finally:
			self.handle.tag_followings.update()#TODO thread?
			self.side_menu.stream_selector.new_tag_followings( self.handle.tag_followings )
			if from_page:
				tag_id = self.get_tag_id(name)
				if tag_id:
					self.stream_pages['followed_tags'].header.tag_followed(tag_id) # let the page know it has been successfully followed
					self.side_menu.stream_selector.del_viewed_tag(name) # remove tag from recent viewed

	# Tags button in side menu
	def on_load_tags_button(self):
		if not self.stream_pages['followed_tags'].is_busy:
			if not self.stream_pages['followed_tags'].stream or self.stream_pages['followed_tags'].current_tag != None:
				self.stream_pages['followed_tags'].current_tag = None
				self.stream_pages['followed_tags'].clear_layout()
				self.new_stream('followed_tags', location='followed_tags.json')

				self.side_menu.stream_selector.deselect_tag()
				self.stream_pages['followed_tags'].header.hide()

	def on_tags_button(self):
		if self.handle:
			frame_visible = not self.side_menu.stream_selector.ui.tags_frame.isHidden()
			if self.handle.tag_followings:
				if not frame_visible:
					# Open tags
					self.side_menu.stream_selector.open_tags()
				else:
					# Close tags
					self.side_menu.stream_selector.close_tags()

			elif not self.stream_pages['followed_tags'].is_busy:
				self.handle.get_tag_followings()
				self.handle.tag_followings_ready.connect( self.on_got_tag_followings )
				self.side_menu.stream_selector.ui.btn_load_tags.click() # Load followed_tags as default
				self.side_menu.stream_selector.ui.btn_load_tags.setChecked(True)

	def get_tag_id(self, name):
		tag_id = None
		for tag in self.handle.tag_followings:
			if name == tag.name:
				tag_id = tag.id
				break
		return tag_id

	# When a tag is clicked.
	def open_tag_page(self, name):
		"""
		param name: Name of the hashtag
		type: str """
		# Check if we already received our followed_tags.
		if not self.handle.tag_followings:
			self.handle.get_tag_followings()
			self.handle.tag_followings_ready.connect( lambda: self.open_tag_page(name) ) # TODO disconnect signal
			self.handle.tag_followings_ready.connect( self.on_got_tag_followings ) # TODO disconnect signal
			return

		# If there already is a thread getting something todo with tags
		if self.stream_pages['followed_tags'].is_busy:
			self.side_menu.stream_selector.deselect_tag()
			return

		# Check if we follow the tag.
		following_tag = None
		for tag in self.handle.tag_followings:
			if name == tag.name:
				following_tag = tag.id
				break

		# Check if different stream that currently is loaded
		if self.stream_pages['followed_tags'].stream and name != self.stream_pages['followed_tags'].current_tag:
			self.stream_pages['followed_tags'].clear_layout()

			# If no thread is already running for tags then get the stream
			self.new_stream('followed_tags', location='tags/{}.json'.format(name))

			# Set current tag on the page.
			self.stream_pages['followed_tags'].set_tag( name, following_tag )

			# Add the tag to the side menu under recent viewed tags if we don't follow it.
			if not following_tag:
				self.side_menu.stream_selector.add_new_viewed_tag(name)
				self.side_menu.stream_selector.open_tags()
		elif not self.stream_pages['followed_tags'].stream:
			self.new_stream('followed_tags', location='tags/{}.json'.format(name))
			self.stream_pages['followed_tags'].set_tag( name, following_tag )

		# Show header (this is hidden on stream with all tags)
		self.stream_pages['followed_tags'].header.show()

		# Open tags page with the tag.
		self.change_stream_page('followed_tags')

	""" Aspects stream """
	def filtered_aspects(self):
		self.stream_pages['aspects'].stop_top_loading()
		self.stream_pages['aspects'].clear_layout()
		self.stream_pages['aspects'].insert_post_widgets()

	def filter_aspects(self):
		""" This is called when aspects refresh icon in side menu is pressed
		Hook to this is made in self.init_pages()

		TODO check if same ids, if so just return """
		cur_selected_ids = self.side_menu.stream_selector.selected_aspect_ids()
		cur_selected_ids = [str(id) for id in cur_selected_ids]

		worker = stream_worker('aspects_filter', aspect_ids_filter=cur_selected_ids, stream=self.stream_pages['aspects'].stream)
		worker.signals.finished.connect( self.filtered_aspects )
		self.handle.stream_pool.start(worker)

		self.stream_pages['aspects'].start_top_loading()

	def on_aspect_button(self):
		if self.handle:
			if self.side_menu.stream_selector.aspects: # if the side menu's stream_selector has list with aspects
				if self.side_menu.stream_selector.ui.aspects_frame.isHidden():
					self.side_menu.stream_selector.open_aspects()
				else:
					self.side_menu.stream_selector.close_aspects()
			elif 'aspects' in self.handle.connection.userdata(): # this should be done once.
				for aspect in self.handle.connection.userdata()['aspects']:
					self.side_menu.stream_selector.add_aspect(aspect['name'], aspect['id'], aspect['selected'])
				self.new_stream('aspects', aspects=True)
				self.on_aspect_button() # when apects inserted in menu, show the Frame

			self.side_menu.stream_selector.selected_aspects

	""" People page functions """
	def people_page_stream_ready(self,guid):
		self.stream_pages['people'].set_stream( self.handle.people[guid].stream )
		self.stream_pages['people'].insert_post_widgets()
		self.stream_pages['people'].scroll_value_changed.connect( self.scroll_changed )

	def people_page_fetch_stream(self,guid):
		print("people_page_fetch_stream() guid: {}".format(guid))
		""" TODO 
			Check if there is already a thread running wich fetches a 
			people stream or people userdata, if so; KILL it :P, we 
			have only 1 people page..
		"""
		if not self.handle.people_stream_pool.activeThreadCount():
			worker = people_worker('fetch_stream', people_obj=self.handle.people[guid])
			worker.signals.new_data.connect( lambda: self.people_page_stream_ready(guid) )
			worker.signals.finished.connect(self.stream_pages['people'].stop_top_loading)
			self.handle.people_stream_pool.start(worker)
			self.stream_pages['people'].start_top_loading()

	def people_page_userdata_ready(self, guid):
		# Set diaspy.people.User object
		self.stream_pages['people'].profile.set_user( self.handle.people[guid] )

		# Update avatar
		worker = img_worker(self.handle.res, url=self.stream_pages['people'].profile._user['profile']['avatar']['large'])
		worker.signals.finished.connect( self.stream_pages['people'].profile.update_avatar )
		self.handle.res.thread_pool.start(worker)

		# Connect btn `load posts` to signal
		self.stream_pages['people'].profile.ui.btn_load_posts.clicked.connect(lambda: self.people_page_fetch_stream(guid))

	def people_page_get_userdata(self, guid):
		self.handle.new_people_obj.disconnect( self.people_page_get_userdata ) # disconnect new_people_obj signal

		worker = people_worker('get_userdata', people_obj=self.handle.people[guid], guid=guid)
		worker.signals.new_data.connect( self.people_page_userdata_ready )
		self.handle.people_pool.start(worker)

	def on_people_event(self, guid=None, diaspora_id=None):
		if not self.handle.connection: return
		print("open people_page {}".format(guid))

		self.side_menu.deselect()
		self.change_stream_page('people')

		if not guid and not diaspora_id:								# when just page needs to be opened
			# Set own profile
			guid = self.handle.connection.userdata()['guid']
			#return

		elif diaspora_id:
			if diaspora_id in self.handle.people_handle_to_guid:
				guid = self.handle.people_handle_to_guid[ diaspora_id ]

		if self.stream_pages['people'].profile._user == None or self.stream_pages['people'].profile._user.guid() != guid:
			""" clear the stream_layout and tag_box """
			self.stream_pages['people'].clear_layout()	# clear posts
			self.stream_pages['people'].profile.reset()	# reset profile
			self.stream_pages['people'].stream = None	# reset stream

			""" disconnect scrollbar from loading more posts """
			try:
				self.stream_pages['people'].scroll_value_changed.disconnect( self.scroll_changed )
			except TypeError:
				pass
			try: # disconnect the `load posts` button
				self.stream_pages['people'].profile.ui.btn_load_posts.clicked.disconnect()
			except TypeError:
				pass # signal was not connected

			""" if it's a different guid get user data handler.people.get_user_data"""
			if guid not in self.handle.people:
				# TODO .. clear avatar, name, diaspora_id, bio etc and disable aspects button and set loading

				worker = people_worker('add_people_obj', connection=self.handle.connection, guid=guid, diaspora_id=diaspora_id)
				worker.signals.new_obj.connect( self.handle.add_people_obj ) # add new people_obj when thread finished
				self.handle.people_pool.start(worker)

				self.handle.new_people_obj.connect( self.people_page_get_userdata ) # when new people_obj connect to self.people_page_get_userdata
			elif self.handle.people_pool.activeThreadCount() == 0: # if people are clicky
				self.people_page_userdata_ready(guid)

	""" Search result pages """
	def on_people_results(self, usersdata, query):
		self.pages['people_results'].set_results(usersdata, query) # set results
		self.change_page('people_results')

	def on_tag_results(self, tags):
		self.pages['tag_results'].set_tags(tags) # set tags
		self.change_page('tag_results')

	""" Connection and login """
	def connect(self):
		self._config['connection']['pod'] = self.ui.line_pod.text()
		self._config['connection']['user'] = self.ui.line_username.text()
		self._config['connection']['pass'] = self.ui.line_password.text()

		# Disable input fields
		self.ui.line_pod.setEnabled(False)
		self.ui.line_username.setEnabled(False)
		self.ui.line_password.setEnabled(False)

		""" Start a thread that tries to make a connection """
		print("REPR SELF HANDLE CONNECTION") # DEBUG
		print(repr(self.handle.connection)) # DEBUG
		if not self.handle.connection:
			self.handle.connect(config=self._config['connection'])
			self.handle.connected.connect(self.connected)
			self.handle.message.connect(self.message)
			self.ui.btn_connect.setEnabled(False)
			self.ui.btn_connect.setText("Connecting..")
		else:
			self.login()

	""" Called when connected """
	def connected(self, state):
		if self.handle.connection:
			print("connected")
			self.login()
			self.ui.btn_connect.setText("Signing in..")
		else:
			self.msg.print_msg("Connection failed", "Could not connect to pod.")

			# TODO make function to enable stuff agian.
			# Enable input fields
			self.ui.line_pod.setEnabled(True)
			self.ui.line_username.setEnabled(True)
			self.ui.line_password.setEnabled(True)

			# Enable connect button
			self.ui.btn_connect.setText("Connect")
			self.ui.btn_connect.setEnabled(True)

	def disconnect(self):
		# TODO if any active threads, KILL them.
		# ...

		# TODO disable tag followings button

		# self.handle
		self.handle.userdata_ready.disconnect()							# disconnect old signal
		self.handle.disconnect()										# set self.handle.res to None

		# header
		self.ui.header.update_icon.clicked.disconnect( self.update_posts )
		self.ui.header.ui.btn_notify.clicked.disconnect( self.on_notifications_button )
		self.ui.header.author_name.clicked.disconnect( self.on_people_event )

		# clear header
		self.ui.header.author_name.set_guid( None )
		self.ui.header.author_name.setToolTip( None )
		self.ui.header.author_name.set_text( "Disconnected" )
		# TODO reset avatar ..

		self.ui.header.set_notification_count( 0 )
		self.ui.header.set_message_count( 0 )

		# menu
		self.side_menu.stream_selector.tag_clicked.disconnect( self.open_tag_page )
		self.side_menu.stream_selector.selected.disconnect( self.change_stream_page )
		self.side_menu.stream_selector.deselected.disconnect( self.leave_page )

		for page_name in self.stream_pages:
			# Disconnect page's signals
			self.stream_pages[page_name].permalink_clicked.connect( self.on_single_post_event )
			try:
				self.stream_pages[page_name].scroll_value_changed.disconnect( self.scroll_changed )
			except TypeError: # when end of stream is reached, it's already disconnected.
				pass

			# clear streams
			self.stream_pages[page_name].reset()

		# TODO clear people page
		# TODO clear contacts page

		# the button itself
		self.ui.btn_connect.clicked.disconnect()
		self.ui.btn_connect.clicked.connect(self.connect)
		self.ui.btn_connect.setText("Connect")
		self.ui.btn_connect.setEnabled(True)

		# Enable input fields
		self.ui.line_pod.setEnabled(True)
		self.ui.line_username.setEnabled(True)
		self.ui.line_password.setEnabled(True)

	def login(self): # move this to handler?
		self.handle.login()
		self.handle.auth_status.connect( self.login_status_change )

	""" Update header """
	def update_header(self): # Should be used one time
		self.ui.header.author_name.set_guid( self.handle.connection.userdata()['guid'] )
		self.ui.header.author_name.set_text( self.handle.connection.userdata()['name'] )
		self.ui.header.author_name.setToolTip( self.handle.connection.userdata()['diaspora_id'] )

		worker = img_worker(self.handle.res, url=self.handle.connection.userdata()['avatar']['small'])
		worker.signals.finished.connect(self.ui.header.update_avatar)

		# TODO create avatar() function in main wich as default returns own avatar
		# This should not be here, nothing to do with header..
		for page_name in self.stream_pages:
			worker.signals.finished.connect( self.stream_pages[page_name].set_own_avatar )
		self.handle.res.thread_pool.start(worker)

		self.ui.header.set_notification_count( self.handle.connection.userdata()['notifications_count'] )
		self.ui.header.set_message_count( self.handle.connection.userdata()['unread_messages_count'] )

		# Set search (we have connection at this point)
		self.ui.header.set_search( self.handle.search )

		# On search hooks
		self.ui.header.request_people_results_page.connect( self.on_people_results )
		self.ui.header.request_tag_results_page.connect( self.on_tag_results )

		# set res for search completion
		self.ui.header.set_res(self.handle.res)

	def get_page_max_width(self, page_name):
		""" TODO create a default page widget that contains scrollArea
		and stream_widget should be a subclass of it. Other pages 
		should use it aswell so we have a standard model for resizing 
		stuff """
		if page_name in self.stream_pages:
			max_width = self.stream_pages[page_name].ui.scrollArea.width()
			max_width -= self.stream_pages[page_name].ui.scrollArea.verticalScrollBar().width()
			max_width -= self.stream_pages[page_name].ui.stream_layout.spacing()
			content_margins = self.stream_pages[page_name].ui.stream_layout.contentsMargins()
		elif page_name in ['contacts']:#, 'people_results', 'tag_results']:
			max_width = self.pages[page_name].ui.scrollArea.width()
			max_width -= self.pages[page_name].ui.scrollArea.verticalScrollBar().width()
			max_width -= self.pages[page_name].ui.contacts_content_layout.spacing()
			content_margins = self.pages[page_name].ui.contacts_content_layout.contentsMargins()
			max_width -= 24

		max_width -= content_margins.left()
		max_width -= content_margins.right()
		return max_width

	def init_pages(self):		
		# Stream selector buttons switch to page
		self.side_menu.stream_selector.selected.connect( self.change_stream_page ) # on enter page
		self.side_menu.stream_selector.deselected.connect( self.leave_page ) # on leave page
		self.side_menu.stream_selector.filter_aspects.connect( self.filter_aspects ) # on aspects filter button in side menu

		# Go to stream page
		self.side_menu.stream_selector.ui.btn_stream.click()

		""" Set max_width use stream page as example else we have to 
		view each page first or maybe there is a function to call so 
		sizes get updated without viewing? """
		# TODO this can be calculated from stream widget itself and if set here it also needs to be set when resized.
		# for now this isn't a problem for pages containing that basicly are a stream widget
		max_width = self.get_page_max_width('stream')

		for page_name in self.stream_pages:
			self.stream_pages[page_name].set_res(self.handle.res)

			self.stream_pages[page_name].set_own_guid( self.handle.connection.userdata()['guid'] )

			self.stream_pages[page_name].set_max_width( max_width )

			# connect author_clicked to open people page
			self.stream_pages[page_name].author_clicked.connect( self.on_people_event )

			# set search object for pages
			self.stream_pages[page_name].set_search( self.handle.search )

			# Post permalinks
			self.stream_pages[page_name].permalink_clicked.connect( self.on_single_post_event )

		# Set res for contacts page
		self.pages['contacts'].set_res(self.handle.res)

		# Max width
		self.pages['contacts'].set_max_width( max_width )
		self.ui.header.set_max_width( max_width ) # header (for search suggestion popup)
		self.pages['tag_results'].set_max_width( max_width ) # tag_results page (for tags box)

		# Search results pages
		self.pages['people_results'].set_res(self.handle.res)
		self.pages['people_results'].user_clicked.connect(self.on_people_event)
		self.pages['people_results'].anchor_clicked.connect(self.on_hyperlink)
		self.pages['tag_results'].tag_clicked.connect(self.open_tag_page)

		# Markdown editor needs aspects for pulisher. TODO use this for other pages that need it.
		self.stream_pages['stream'].markdown_editor.set_aspects( self.handle.connection.userdata()['aspects'] )
		self.stream_pages['stream'].markdown_editor.set_publisher( self.handle.publisher ) # TODO this is a try-out

		# People page's profile needs list of aspects too
		self.stream_pages['people'].profile.set_aspects( self.handle.connection.userdata()['aspects'] )

		# Contacts page like apsects very much nom nom nom
		self.pages['contacts'].set_aspects( self.handle.connection.userdata()['aspects'] )
		self.pages['contacts'].user_clicked.connect( self.on_people_event )

		# When aspects get updated
		self.pages['contacts'].aspects_updated.connect( self.stream_pages['stream'].markdown_editor.publisher.aspects_button.refresh )
		self.pages['contacts'].aspects_updated.connect( self.stream_pages['people'].profile.update_aspects_button ) # TODO this doesn't work..
		self.pages['contacts'].contacts_updated.connect( self.stream_pages['people'].profile.update_aspects_button )

		#self.stream_pages['people'].profile.aspects_updated.connect( self.pages['contacts'].refresh_contact_aspects )
		self.stream_pages['people'].profile.added_user_to_aspect.connect( self.pages['contacts'].outside_contact_added )
		self.stream_pages['people'].profile.removed_user_from_aspect.connect( self.pages['contacts'].outside_contact_removed )

		# Load more posts TODO move this after stream inited
		self.stream_pages['stream'].scroll_value_changed.connect( self.scroll_changed )
		self.stream_pages['aspects'].scroll_value_changed.connect( self.scroll_changed )
		self.stream_pages['mentions'].scroll_value_changed.connect( self.scroll_changed )
		self.stream_pages['activity'].scroll_value_changed.connect( self.scroll_changed )
		self.stream_pages['public'].scroll_value_changed.connect( self.scroll_changed )
		self.stream_pages['followed_tags'].scroll_value_changed.connect( self.scroll_changed )

		# (un)Follow tag buttons
		self.stream_pages['followed_tags'].header.request_follow_tag.connect( lambda tag_name: self.on_new_tag(tag_name, from_page=True) )
		self.stream_pages['followed_tags'].header.request_unfollow_tag.connect( self.del_tag_from_page )

		# Tags
		self.side_menu.stream_selector.tag_clicked.connect( self.open_tag_page )
		self.side_menu.stream_selector.ui.btn_load_tags.clicked.connect( self.on_load_tags_button ) # load all tags in stream
		self.pages['contacts'].tag_clicked.connect( self.open_tag_page )
		self.stream_pages['people'].profile.tag_clicked.connect( self.open_tag_page )

		# Get stream
		self.new_stream('stream') # the default stream

	def login_status_change(self, status):
		if status:
			print("Signed in succesfully")
			# Set config verified status to True
			if self._config['connection']['verified'] == "False": self.save_connection_settings("True")

			# Set connect button on settings page to disconnect and remove signal
			self.ui.btn_connect.setText("Logout")
			self.ui.btn_connect.clicked.disconnect()
			self.ui.btn_connect.clicked.connect(self.disconnect)

			# header hooks
			self.ui.header.ui.btn_notify.clicked.connect( self.on_notifications_button )
			self.ui.header.update_icon.clicked.connect( self.update_posts )
			self.ui.header.author_name.clicked.connect( self.on_people_event ) # own name clicked

			# get user data
			self.handle.get_user_data()
			self.handle.userdata_ready.connect( self.update_header )
			self.handle.userdata_ready.connect( self.init_pages )

			# get contacts
			"""The self.handle.contacts (diaspy.people.Contacts) object
			is used to add and remove users from aspects on people page.
			It'll also be used for the contact page to manage aspects
			and contacts.

			Do this here and now 'cause we now assume to have 
			self.handle.connection at this point"""
			self.handle.set_contacts() # TODO <-- thread?
			self.stream_pages['people'].profile.set_contacts( self.handle.contacts )
			# set contacts on contacts page
			self.pages['contacts'].set_contacts( self.handle.contacts )

		else:
			self.msg.print_msg("Login failed", "Username or password incorrect for this pod." )
			self.ui.btn_connect.setText("Connect")

			# Enable input fields
			self.ui.line_pod.setEnabled(True)
			self.ui.line_username.setEnabled(True)
			self.ui.line_password.setEnabled(True)
		self.ui.btn_connect.setEnabled(True)

	""" Settings """
	def save_connection_settings(self, verified="False"):
		self._config['connection']['pod'] = self.ui.line_pod.text()
		self._config['connection']['user'] = self.ui.line_username.text()
		self._config['connection']['pass'] = self.ui.line_password.text()
		if verified != "False":
			self._config['connection']['verified'] = "{}".format(verified)
		self.msg.print_msg("Save settings", "Connection settings saved.")

	def save_settings(self):
		with open(self._config_file, 'w') as configfile:
			self._config.write(configfile)

	""" Podlist (settings page) """
	def set_pod(self, pod):
		pod_url = self.pl.podlist.tbl_podlist.item(pod.row(), 0).text()
		self.ui.line_pod.setText( pod_url )
		self._config['connection']['pod'] = pod_url
		with open(self._config_file , 'w') as configfile:
			self._config.write(configfile)

	def handle_podlist_state(self, obj):
		if self.ui.lo_podlist.count() > 0:
			self.ui.btn_open_podlist.setText("Open podlist")
			self.clear_layout(self.ui.lo_podlist)
			self.pl = None
		else:
			if self.pl is None:
				self.pl = podlist()
			self.ui.lo_podlist.addWidget(self.pl)
			self.pl.podlist.tbl_podlist.doubleClicked.connect(self.set_pod)
			self.ui.btn_open_podlist.setText("Close podlist")

	def remove_from_list(self, list, name):
		return list.remove(name)

	def more_posts_check(self, name='stream'):
		if not self.stream_pages[name].new_guids:
			if self.stream_pages[name].more_fail_count == self.stream_pages[name].more_fail_count_max:
				self.stream_pages[name].scroll_value_changed.disconnect()
				self.msg.print_msg("Stream", "Reached end of the stream, not loading anymore posts.")
			else:
				self.stream_pages[name].more_fail_count +=1

	def more_posts(self, name='stream'):
		if name not in self.thread_list['stream']['more']:
			self.thread_list['stream']['more'].append(name) # add to thread_list wich is used for checking if same is not already running
			worker = stream_worker('more', name=name, stream=self.stream_pages[name].stream)
			worker.signals.new_data.connect( self.stream_pages[name].insert_post_widgets )
			worker.signals.new_data.connect( lambda: self.more_posts_check(name) ) # Check if no more posts.
			worker.signals.finished.connect( self.stream_pages[name].stop_bottom_loading )
			worker.signals.finished.connect( lambda: self.remove_from_list(self.thread_list['stream']['more'], name) )
			self.handle.stream_pool.start(worker)
			self.stream_pages[name].start_bottom_loading()

	def update_posts(self, name='stream'):
		try:
			name=self.ui.stackedWidget.currentWidget().name
		except AttributeError:
			return # I guess it's the settings page.
		if name not in self.thread_list['stream']['update']:
			if self.stream_pages[name].stream: # if stream for that page exists
				self.thread_list['stream']['update'].append(name) # add to thread_list wich is used for checking if same is not already running

				worker = stream_worker('update', name=name, stream=self.stream_pages[name].stream)
				worker.signals.new_data.connect( self.stream_pages[name].insert_post_widgets )
				worker.signals.finished.connect(self.stream_pages[name].stop_top_loading)
				worker.signals.finished.connect( self.ui.header.update_icon.pause )
				worker.signals.finished.connect(lambda: self.remove_from_list(self.thread_list['stream']['update'], name))
				self.handle.stream_pool.start(worker)

				self.ui.header.update_icon.start()
				self.stream_pages[name].start_top_loading()
			else:
				pass
				# TODO get stream if it previous failed.

	""" Will call more_posts() when no thread is already running
	and scroll bar is over 95 percent down 
	TODO check if scrollbar != 10px or so """
	def scroll_changed(self, page_name, value):
		if self.stream_pages[page_name].stream: # if stream for that page exists

			""" load more posts """
			one_percent =  self.stream_pages[page_name].ui.scrollArea.verticalScrollBar().maximum() / 100
			if one_percent: # don't want ZeroDivisionError
				if round(value / one_percent) > 95:
					self.more_posts(name=page_name)

			""" post visable """
			# ..

	""" On resize event """
	def resizeEvent(self, event):
		self.resized.emit()
		QtWidgets.QMainWindow.resizeEvent(self, event)

		# Save settings
		# width + height
		size = self.size()
		self._config['general']['width'] = str(size.width())
		self._config['general']['height'] = str(size.height())

		# TODO maximized state is not detected right this way..
		# maximized state
		if self.isMaximized(): self._config['general']['maximized'] = 'True'
		else: self._config['general']['maximized'] = 'False'

		# fullscreen state
		if self.isFullScreen(): self._config['general']['fullscreen'] = 'True'
		else: self._config['general']['fullscreen'] = 'False'

		self.save_settings()

	def resize_event(self, post_guid=None, page_name=None, layout='default', extract=0): # TODO clean this mess
		self.msg.update_size( size=self.size(), margins=self.ui.verticalLayout.contentsMargins() ) # TODO this should not be here, or beter to rename this function to post_resized oid
		self.notifications.update_size( size=self.size(), margins=self.ui.verticalLayout.contentsMargins() )

		# Get max_width of currently viewed page
		try:
			name=self.ui.stackedWidget.currentWidget().name
		except AttributeError: # if failed fallback on stream page
			name = 'stream'
		max_width = self.get_page_max_width(name)

		if page_name:
			layout = self.stream_pages[page_name].ui.stream_layout
			self.stream_pages[page_name].set_max_width( max_width )
			print("using page_name: {}".format(page_name))

		elif layout == 'default': # resize all items on all stream pages
			self.pages['contacts'].set_max_width( max_width ) # contacts page, TODO this should be on different place, 'cause it's called much to often now.
			self.ui.header.set_max_width( max_width ) # header (for search suggestion popup)
			self.pages['tag_results'].set_max_width( max_width ) # tag_results page (for tags box)

			for page_name in self.stream_pages:
				self.resize_event( page_name=page_name )
			return

		else:
			max_width = layout.parentWidget().sizeHint().width()
			print("using layout max_width NOT FROM reshare: {}".format(max_width))

		if post_guid != None:
			print("trying to set max width")
			layout.post_widgets[post_guid].set_max_width( max_width )
			return

		for i in range(layout.count()):
			item = layout.itemAt(i)
			if not item:
				continue
			w = item.widget()

			if w:
				if type(w) is post_widget:
					w.set_max_width( max_width )

""" TMP DEBUG
copied function from https://pymotw.com/2/sys/tracing.html

DON'T SET TO TRUE ON OLDER SYSTEMS OR IT WILL BE REALLY SLOW!
"""
#__DEBUG=True
__DEBUG=False
if __DEBUG:
	import sys
	def trace_calls(frame, event, arg):
		if event != 'call':
			return
		co = frame.f_code
		func_name = co.co_name
		if func_name == 'write':
			# Ignore write() calls from print statements
			return
		func_line_no = frame.f_lineno
		func_filename = co.co_filename
		caller = frame.f_back
		caller_line_no = caller.f_lineno
		caller_filename = caller.f_code.co_filename
		print('Call to {} on line {} of {} from line {} of {}'.format(func_name, func_line_no, func_filename, caller_line_no, caller_filename))
		return

	sys.settrace(trace_calls)

if __name__ == "__main__":
	import sys
	app = QtWidgets.QApplication(sys.argv)
	w = Window()
	w.show()
	sys.exit(app.exec_())
