"""
This file is part of LionsTooth.

LionsTooth is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

LionsTooth is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with LionsTooth.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import pyqtSignal

from design.ui_contacts import Ui_contacts
from design.ui_contact import Ui_contact
#from design.ui_hover_card import Ui_hover_card

from widgets.aspects_button import aspect_item
from widgets.shared.author_name import author_name
from pages.people import profile_aspects_button
from classes.resolver import img_worker

from widgets.shared.tags_box import tags_box

class contact(QtWidgets.QFrame):
	user_clicked = pyqtSignal(str, str)# guid, diaspora_id
	add_to_aspect = pyqtSignal(int, int, int) # user_id, contact_id, aspect id
	remove_from_aspect = pyqtSignal(int, int) # user_id, aspect id
	tag_clicked = pyqtSignal(str) # tag_name
	def __init__(self, user, aspects, max_width=100, parent=None):
		QtWidgets.QFrame.__init__(self, parent)
		
		"""
		param user: User contains the data we are going to set
		type user: diaspy.people.User object
		"""
		self.user = user
		self.ui = Ui_contact()
		self.ui.setupUi(self)

		self.max_width = max_width
		
		self.user_id = self.user.id()

		self.ui.name = author_name(guid=self.user.data['person']['guid'], diaspora_id=self.user.data['person']['diaspora_id'])
		self.ui.name.clicked.connect( self.user_clicked )
		self.ui.content_layout.insertWidget(0, self.ui.name)

		""" Aspects button """
		self._apsects = aspects # this will have diaspy objects
		self.aspects = [] # this will be for the layout
		self.compile_aspects()

		self.ui.aspects_button = profile_aspects_button(aspects=self.aspects)
		self.ui.contact_layout.addWidget( self.ui.aspects_button )

		""" aspect button hooks """
		self.ui.aspects_button.remove_from_aspect.connect( self.emit_remove_from_aspect )
		self.ui.aspects_button.add_to_aspect.connect( self.emit_add_to_aspect )

		""" tags box """
		self.margins = 100+12+12+(4*8) # This is without the aspect_button, useless to request it's size in init.
		
		self.tags_box = tags_box(max_width=self.max_width-self.margins)
		self.tags_box.tag_clicked.connect( self.tag_clicked )
		self.ui.content_layout.addWidget( self.tags_box )
		
		self.set_data()

	def calc_margins(self): # TODO make dynamic
		self.margins = (100+12+12+(4*8)+self.ui.aspects_button.size().width())

	def set_max_width(self, max_width):
		self.max_width = max_width
		self.calc_margins()
		
		self.tags_box.set_max_width( max_width-self.margins) # avatar, layout_left_margin, layout_right_margin + 3* layout_spacing, aspects_button width

	def outside_contact_added(self, new_aspect_membership):
		""" When a contact is added from an aspect ouside of this object
		then use this function to add it to this page too. """
		new_aspect_membership.update( { 'selected': True } )
		self.user.data['aspect_memberships'].append(new_aspect_membership)
		self.compile_aspects()
		self.ui.aspects_button.set_aspects( self.aspects )
		
		self.ui.aspects_button.refresh()

	def outside_contact_removed(self, aspect_id):
		""" When a contact is removed from an aspect outside of this 
		object then use this function to update this page. """
		for a in self.user.data['aspect_memberships']:
			if a['aspect']['id'] == aspect_id:
				self.user.data['aspect_memberships'].remove(a)
				self.compile_aspects()
				self.ui.aspects_button.set_aspects( self.aspects )
				
				self.ui.aspects_button.refresh()
				break

	def compile_aspects(self):
		"""Here we create a list with aspect_id's that the user 
		currently has a membership.
		
		Important is that self._aspects and 
		self.user.data['aspect_memberships'] are up-to-date.
		
		We take a list with a dict called self._aspects and iter over it
		if the key id from self._aspects is in the previously created
		list with memberships then set it to select, otherwise set 
		selected False"""
		if self.user.data['aspect_memberships']:
			ids = [self.user.data['aspect_memberships'][i]['aspect']['id'] for i in range(len(self.user.data['aspect_memberships']))]
		else:
			ids = []

		self.aspects = []
		for a in self._apsects:
			selected = False
			
			# If user is member of the aspect select it
			if int(a['id']) in ids:
				selected = True

			self.aspects.append( { 
				'id'		: a['id'],
				'name'		: a['name'],
				'selected'	: selected } )

	def emit_remove_from_aspect(self, aspect_id):
		# Sort of a request to contacts page to handle it
		self.remove_from_aspect.emit( self.user.data['id'], aspect_id ) # contact id

	def emit_add_to_aspect(self, aspect_id):
		# Sort of a request to contacts page to handle it
		self.add_to_aspect.emit( self.user.data['person_id'], self.user.data['id'], aspect_id ) # user id

	def set_avatar(self, path):
		if path:
			pixmap = QtGui.QPixmap(path)
			if pixmap:
				self.ui.avatar.setMaximumSize(QtCore.QSize(100, 100))
				self.ui.avatar.setMinimumSize(QtCore.QSize(100, 100))
				self.ui.avatar.setPixmap(pixmap)

	def set_data(self):
		self.ui.name.set_text( self.user.data['person']['name'] )
		self.ui.diaspora_id.setText( self.user.data['person']['diaspora_id'] )
		for tag_name in self.user.data['person']['profile']['tags']:
			self.tags_box.add_tag( tag_name )

		self.tags_box.set_max_width( self.max_width - self.margins )
		self.tags_box.change_layout()
		

class aspect_view_button(QtWidgets.QFrame):
	clicked = pyqtSignal(str, bool) # str: id; bool: selected or deselected (True or False)
	delete_clicked = pyqtSignal(str) # str: id
	def __init__(self, name, id, count=0, parent=None):
		QtWidgets.QFrame.__init__(self, parent)
		
		self.name = name
		self.count = count
		self.id = str(id) # str 'cause it also can be 'all', 'contacts' pr 'only_sharing'
		self.__selected = False
		
		self.setStyleSheet("")
		self.horizontalLayout = QtWidgets.QHBoxLayout(self)
		self.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))

		""" aspect count """
		self.aspect_count = QtWidgets.QLabel(self)
		sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
		
		self.aspect_count.setMinimumWidth( 50 )
		self.aspect_count.setMaximumWidth( 50 )
		
		self.aspect_count.setSizePolicy(sizePolicy)
		self.set_count( count )
		self.horizontalLayout.addWidget(self.aspect_count)
		
		""" aspect name """
		self.aspect_name = QtWidgets.QLabel(self)
		self.aspect_name.setStyleSheet("text-align: left;\n"
"margin-left:5px;")
		self.aspect_name.setText(name)
		self.aspect_name.setSizePolicy(sizePolicy)
		self.horizontalLayout.addWidget(self.aspect_name, 0, QtCore.Qt.AlignLeft)

		""" delete aspect button """
		if self.id not in ['all', 'contacts', 'only_sharing']:
			self.btn_delete = QtWidgets.QPushButton(self)
			self.btn_delete.setText("X")
			self.btn_delete.setFlat(True)
			self.btn_delete.setSizePolicy(sizePolicy)
			self.btn_delete.setToolTip("Click here to delete this whole aspect.")
			
			sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
			self.btn_delete.setSizePolicy(sizePolicy)
			
			self.horizontalLayout.addWidget(self.btn_delete, 0, QtCore.Qt.AlignRight)

			self.btn_delete.clicked.connect( self._delete )

	def _delete(self):
		self.delete_clicked.emit( self.id )

	def mousePressEvent(self, event): # TODO detect if click is left
		if self.__selected: self._set_selected(False)
		else: self._set_selected(True)
		self.clicked.emit( self.id, self.__selected )
		QtWidgets.QFrame.mousePressEvent(self, event)

	def set_count(self, count):
		self.count = count
		self.aspect_count.setText( "[{}]".format(count) )
	
	def plus_one(self):
		self.count += 1
		self.aspect_count.setText( "[{}]".format(self.count) )
	
	def min_one(self):
		self.count -= 1
		self.aspect_count.setText( "[{}]".format(self.count) )
	
	def _set_selected(self, bool):
		if bool:
			self.__selected = True
			self.setStyleSheet("background-color: #446241;") # selected
		else:
			self.__selected = False
			self.setStyleSheet("background-color: #151e14;") # normal

class add_aspect_dialog(QtWidgets.QDialog):
	def __init__(self, parent=None):
		QtWidgets.QDialog.__init__(self, parent)

		""" layout """
		self.layout = QtWidgets.QVBoxLayout(self)

		""" header """
		self.title = QtWidgets.QLabel(self)
		self.title.setText("<h1>Add new aspect</h1>")
		self.layout.addWidget( self.title )

		""" name: input field """
		self.input_layout = QtWidgets.QHBoxLayout(self)

		self.lbl_name = QtWidgets.QLabel(self)
		self.lbl_name.setText("Name")
		self.input_layout.addWidget( self.lbl_name )

		self.input = QtWidgets.QLineEdit(self)
		self.input.setPlaceholderText("enter a name for your new aspect..")
		self.input_layout.addWidget( self.input )

		self.layout.addItem( self.input_layout )

		""" checkbox: Make contacts in this aspect visible for each """
		self.checkbox = QtWidgets.QCheckBox(self)
		self.checkbox.setText("Make contacts in this aspect visible for each")
		self.layout.addWidget( self.checkbox )

		""" buttons """
		self.button_layout = QtWidgets.QHBoxLayout(self)

		self.btn_cancel = QtWidgets.QPushButton(self)
		self.btn_cancel.setText("Cancel")
		self.button_layout.addWidget( self.btn_cancel )

		self.btn_add = QtWidgets.QPushButton(self)
		self.btn_add.setText("Add")
		self.button_layout.addWidget( self.btn_add )
		
		self.layout.addItem( self.button_layout )

		""" set layout """
		self.setLayout(self.layout)
		
		
		self.setStyleSheet("background-color: #151e14;")
		self.setWindowTitle("Add new aspect")
		self.setFixedHeight(self.sizeHint().height())
		
		""" hooks """
		self.btn_cancel.clicked.connect( self.cancel )
		self.btn_add.clicked.connect( self.add )
		
		self.exec_()

	def add(self):
		if self.input.text(): # TODO some checks if aspect name is valid
			self.close()
	
	def cancel(self):
		self.close()

class contacts(QtWidgets.QWidget):
	user_clicked = pyqtSignal(str, str)# guid, diaspora_id
	contacts_updated = pyqtSignal() # TODO so other objects know that they need to refresh contact data
	aspects_updated = pyqtSignal() # TODO so other objects know that they need to refresh aspect data
	tag_clicked = pyqtSignal(str) # tag_name
	def __init__(self, contacts=None, aspects=None, parent=None):
		QtWidgets.QWidget.__init__(self, parent)
		self.ui = Ui_contacts()
		self.ui.setupUi(self)
		
		self.name = 'contacts'
		
		self._res = None # resource for images
		self._max_width = 0 # so for only used for tags box
		
		self._selected_aspect = None
		self._aspect_view_buttons = {} # key: id; value: widget
		self._contact_widgets = {} # key: contact_id; value: contact widget
		self.aspect_contact_ids = {} # key: aspect_id; value: list with contact_id's that are in that aspect

		self._contacts = contacts # diaspy.people.Contacts
		self._aspects = aspects # diaspy.connection.getUserData()['aspects']
		self._alt_aspects = {} # { id : [name, count] } for the layout
		
		self.ui.btn_add_aspect.clicked.connect( self.add_aspect )
		self.aspects_updated.connect( self.refresh_contact_aspects )

	def set_max_width(self, max_width):
		#print("PAGES -> CONTACTS # SET_MAX_WIDTH : {}".format(max_width))#DEBUG
		if max_width:
			self._max_width = max_width -(50+12+12) # TODO -50 for the side menu, 2*12 margins
			
			for contact_id in self._contact_widgets:
				self._contact_widgets[contact_id].set_max_width( max_width )

	def set_res(self, res):
		self._res = res

	def outside_contact_added(self, contact_id, user_id, aspect_membership):
		# TODO add contact if it doesn't exist yet.
		if contact_id not in self._contact_widgets: return

		self._contact_widgets[contact_id].outside_contact_added( aspect_membership )
		
		# If this is the users first aspect then update only sharing to -1
		if len(self._contact_widgets[contact_id].user.data['aspect_memberships']) == 1:
			self._aspect_view_buttons['only_sharing'].min_one()
			self.aspect_contact_ids['only_sharing'].remove( contact_id ) # keep filter list up to date
		
		aspect_id = str(aspect_membership['aspect']['id'])
		
		# keep filter list up to date
		if aspect_id in self.aspect_contact_ids:
			self.aspect_contact_ids[aspect_id].append( contact_id )
		else:
			self.aspect_contact_ids.update( { aspect_id: [contact_id] } )

		self._aspect_view_buttons[aspect_id].plus_one()

	def outside_contact_removed(self, contact_id, user_id, aspect_id):
		self._contact_widgets[contact_id].outside_contact_removed( aspect_id )
		
		# If user is in no aspects anymore then update only sharing to +1
		if not len(self._contact_widgets[contact_id].user.data['aspect_memberships']):
			self._aspect_view_buttons['only_sharing'].plus_one()
			self.aspect_contact_ids['only_sharing'].append( contact_id ) # keep filter list up to date
		
		self.aspect_contact_ids[str(aspect_id)].remove( contact_id ) # keep filter list up to date
		self._aspect_view_buttons[str(aspect_id)].min_one()

	def add_to_aspect(self, user_id, contact_id, aspect_id):
		if self._contacts:
			self._contacts.add(user_id, [aspect_id]) # add excepts list for aspects_ids
			self._aspect_view_buttons[str(aspect_id)].plus_one()# update count
			
			# If this is the users first aspect then update only sharing to -1
			if len(self._contact_widgets[contact_id].user.data['aspect_memberships']) == 1:
				self._aspect_view_buttons['only_sharing'].min_one()
				self.aspect_contact_ids['only_sharing'].remove( contact_id ) # keep filter list up to date
			
			# keep filter list up to date
			if str(aspect_id) in self.aspect_contact_ids:
				self.aspect_contact_ids[str(aspect_id)].append( contact_id )
			else:
				self.aspect_contact_ids.update( { str(aspect_id): [contact_id] } )
			
			""" update self.aspect_contact_ids """
			if str(aspect_id) not in self.aspect_contact_ids:
				self.aspect_contact_ids.update( { str(aspect_id): [contact_id] } )
			else:
				self.aspect_contact_ids[str(aspect_id)].append( contact_id )
			
			self.contacts_updated.emit()

	def remove_from_aspect(self, contact_id, aspect_id):
		# TODO 
		# ..
		user = None
		for contact in self._contacts:
			if contact.id() == contact_id:
				self._contacts.remove(contact, [aspect_id])
				break
		
		# If user is in no aspects anymore then update only sharing to +1
		if not len(self._contact_widgets[contact_id].user.data['aspect_memberships']):
			self._aspect_view_buttons['only_sharing'].plus_one()
			self.aspect_contact_ids['only_sharing'].append( contact_id ) # keep filter list up to date

		self._aspect_view_buttons[str(aspect_id)].min_one()# update count
		
		# update self.aspect_contact_ids, needed for filter
		self.aspect_contact_ids[str(aspect_id)].remove( contact_id )

	def add_aspect_membership(self, contact_id, new_aspect_membership): # to contact
		if new_aspect_membership:
			self._contact_widgets[contact_id].user.data['aspect_memberships'].append( new_aspect_membership )
			self._contact_widgets[contact_id].compile_aspects()
			# <-- TODO why does it needs to be set again for some reason, I wasn't aware it's a copy.
			self._contact_widgets[contact_id].ui.aspects_button.set_aspects( self._contact_widgets[contact_id].aspects )
			self._contact_widgets[contact_id].ui.aspects_button.refresh()

	def remove_aspect_membership(self, contact_id, aspect_id):
		for membership in self._contact_widgets[contact_id].user.data['aspect_memberships']:
			if membership['id'] == aspect_id:
				self._contact_widgets[contact_id].user.data['aspect_memberships'].remove( membership )
				self._contact_widgets[contact_id].compile_aspects()
				# <-- TODO why does it needs to be set again for some reason, I wasn't aware it's a copy.
				self._contact_widgets[contact_id].ui.aspects_button.set_aspects( self._contact_widgets[contact_id].aspects )
				self._contact_widgets[contact_id].ui.aspects_button.refresh()
				break

	def refresh_contact_aspects(self):
		for contact_id in self._contact_widgets:
			#print("__refresh_contact_aspects button") #DEBUG
			self._contact_widgets[contact_id]._apsects = self._aspects # TODO is this needed?
			self._contact_widgets[contact_id].compile_aspects()

			# <-- TODO why does it needs to be set again for some reason, I wasn't aware it's a copy.
			self._contact_widgets[contact_id].ui.aspects_button.set_aspects( self._contact_widgets[contact_id].aspects )
			self._contact_widgets[contact_id].ui.aspects_button.refresh()

	def add_aspect_dialog(self):
		dialog = add_aspect_dialog(self)
		return dialog.input.text(), dialog.checkbox.checkState()

	def add_aspect(self):
		name, aspect_contacts_visible = self.add_aspect_dialog()
		"""
		TODO markdown_editor's aspects don't get updated
		"""
		if name:
			new_aspect = self._contacts.addAspect( name, visible=aspect_contacts_visible)
			if new_aspect:
				#print(new_aspect)#DEBUG
				

				str_id = str(new_aspect['id'])
				self._alt_aspects.update( { str_id: [new_aspect['name'], 0] } )
				self._aspect_view_buttons.update( { str_id: aspect_view_button(new_aspect['name'], str_id, 0) } )
				#self._aspects.append(new_aspect) # done in diaspy now
				self.ui.aspects_layout.addWidget( self._aspect_view_buttons[str_id] )

				""" hooks """
				self._aspect_view_buttons[str_id].clicked.connect( self.toggle_aspect ) # toggle
				self._aspect_view_buttons[str_id].delete_clicked.connect( self.delete_aspect ) # delete
				
				self.aspects_updated.emit() # let other objects that make use of the same aspects object know it's changed.
				
				# Add to self.aspect_contact_ids
				self.aspect_contact_ids.update( { str_id: [] } )

	def confirm(self, title, quesion):
		answer = QtWidgets.QMessageBox.question(self, title, quesion, QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.No)
		if answer == QtWidgets.QMessageBox.Yes:
			return True
		return False

	def delete_aspect(self, id):
		if self.confirm("Delete aspect", "Are you sure you want to <b>delete</b> this aspect?"):
			#print(id)#DEBUG
			self._contacts.deleteAspect( id )
			self._aspect_view_buttons[id].hide()
			
			self._alt_aspects.pop( id )
			self._aspect_view_buttons.pop( id )
			i = 0
			for a in self._aspects:
				if str(a['id']) == id:
					self._aspects.pop(i)
					break
				i +=1
			
			# Delete from self.aspect_contact_ids
			self.aspect_contact_ids.update( { str_id: [] } )

			self.aspects_updated.emit() # let other objects that make use of the same aspects object know it's changed.

	def set_contacts(self, contacts): # only sets the object, not sure if there are contacts fetched or not.
		if contacts:
			if not contacts.contacts:
				contacts.contacts = contacts.get(set='all') # TODO thread
				
			self._contacts = contacts
			if self._aspects:
				self.process()

	def insert_contacts(self):
		for user in self._contacts:
			user_id = user.id()
			if user_id not in self._contact_widgets:
				self._contact_widgets.update( { user_id: contact(user, self._aspects, max_width=self._max_width) } )
				self.ui.contacts_layout.addWidget( self._contact_widgets[user_id] )
				
				""" hooks """
				self._contact_widgets[user_id].user_clicked.connect( self.user_clicked )
				self._contact_widgets[user_id].add_to_aspect.connect( self.add_to_aspect )
				self._contact_widgets[user_id].remove_from_aspect.connect( self.remove_from_aspect )
				self._contact_widgets[user_id].tag_clicked.connect( self.tag_clicked )
				
				""" avatar thread """
				avatar_worker = img_worker(self._res, url=user.data['person']['profile']['avatar']['medium'])
				avatar_worker.signals.finished.connect( self._contact_widgets[user_id].set_avatar )
				self._res.thread_pool.start(avatar_worker)

	def process(self):
		if self._contacts and self._aspects:
			self.process_contact_aspects()
			self.insert_contacts()

	def set_aspects(self, aspects):
		self._aspects = aspects
		if self._contacts:
			self.process()
	"""
	[{
			"id":123, 													# contact id
			"person_id":123,
			"aspect_memberships":[
				{
					"id":123,											# aspect membership id
					"aspect":{ "id":1000, "name":"Friends" }
				},{
					"id":456,
					"aspect":{ "id":1001, "name":"test2" }
				},{
					"id":789,
					"aspect":{"id":1002,"name":"testing"}
				}],
			"person":{
				"id":123,
				"guid":"0123456789abcdef0123456789abcdef",
				"name":"Test user",
				"diaspora_id":"test@test.test",
				"relationship":"mutual",
				"block":false,
				"is_own_profile":false,
				"show_profile_info":true,
				"profile":{
					"id":123,
					"searchable":false,
					"avatar":{
						"small":"/assets/user/default.png",
						"medium":"/assets/user/default.png",
						"large":"/assets/user/default.png"},
					"tags":[],
					"bio":"",
					"birthday":null,
					"gender":null,
					"location":""
				}
			}
		},{"""
	
	def toggle_aspect(self, id, bool):
		if bool:
			if self._selected_aspect:
				self._aspect_view_buttons[ self._selected_aspect ]._set_selected(False)
			
			self._aspect_view_buttons[ id ]._set_selected(True)
			self._selected_aspect = id
			self.filter_contacts_by_aspect( id )
			
	
	def filter_contacts_by_aspect(self, aspect_id):
		for contact_id in self._contact_widgets:
			
			if aspect_id == 'all': # show all
				self._contact_widgets[contact_id].show()

			elif aspect_id == 'contacts': # filter by contacts
				is_contact = False
				for a_id in self.aspect_contact_ids:
					if a_id == 'only_sharing': continue
					if contact_id in self.aspect_contact_ids[a_id]:
						is_contact = True
						break

				if is_contact: self._contact_widgets[contact_id].show()
				else: self._contact_widgets[contact_id].hide()

			elif aspect_id in self.aspect_contact_ids: # filter by aspect_id and 'only_sharing'
				if contact_id in self.aspect_contact_ids[aspect_id]:
					self._contact_widgets[contact_id].show()
				else:
					self._contact_widgets[contact_id].hide()
			
			else: # no contacts, hide all
				self._contact_widgets[contact_id].hide()
	
	def process_contact_aspects(self):
		if self._contacts:
			aspects = {}

			for aspect in self._aspects:
				aspects.update( { str(aspect['id']): [aspect['name'], 0, aspect['selected']] } )

			contact_count = 0
			all_contact_count = len(self._contacts.contacts)
			for contact in self._contacts:
				if contact.data['aspect_memberships']:
					contact_count +=1
					for membership in contact.data['aspect_memberships']:
						id = str(membership['aspect']['id'])
						
						""" keep a list so we know what contacts 
						should be visible and what not if filtered on 
						aspect. """
						if id not in self.aspect_contact_ids:
							self.aspect_contact_ids.update( { id: [contact.id()] } )
						else:
							self.aspect_contact_ids[id].append( contact.id() )

						""" """
						name = membership['aspect']['name']
						if id not in aspects:
							aspects.update( { id: [name, 1] } )
						else: aspects[id][1] +=1
					self._alt_aspects = aspects
				else:
					if 'only_sharing' in self.aspect_contact_ids:
						self.aspect_contact_ids['only_sharing'].append( contact.id() )
					else:
						self.aspect_contact_ids.update( { 'only_sharing': [contact.id()] } )

			""" add aspect buttons """
			for id in aspects:
				str_id = str(id)
				self._aspect_view_buttons.update( { str_id: aspect_view_button(aspects[id][0], str_id, aspects[id][1]) } )
				
				""" hooks """
				self._aspect_view_buttons[str_id].delete_clicked.connect( self.delete_aspect ) # delete hook
				self._aspect_view_buttons[str_id].clicked.connect( self.toggle_aspect )
				
				""" add widget """
				self.ui.aspects_layout.addWidget( self._aspect_view_buttons[str_id] )

			""" add default options like 'All contacts', 'My contacts',
			'Contacts who only share with me' """
			self._aspect_view_buttons.update( { 'all': aspect_view_button("<b>All contacts</b>", 'all', all_contact_count) } )
			self._aspect_view_buttons.update( { 'contacts': aspect_view_button("<b>My contacts</b>", 'contacts', contact_count) } )
			self._aspect_view_buttons.update( { 'only_sharing': aspect_view_button("<b>Only sharing with me</b>", 'only_sharing', all_contact_count-contact_count) } )

			self._aspect_view_buttons['all']._set_selected(True)
			self._selected_aspect = 'all'
			
			self._aspect_view_buttons['all'].clicked.connect( self.toggle_aspect )
			self._aspect_view_buttons['contacts'].clicked.connect( self.toggle_aspect )
			self._aspect_view_buttons['only_sharing'].clicked.connect( self.toggle_aspect )

			self.ui.aspects_layout.insertWidget(1, self._aspect_view_buttons['all'] )
			self.ui.aspects_layout.insertWidget(2, self._aspect_view_buttons['contacts'] )
			self.ui.aspects_layout.addWidget(self._aspect_view_buttons['only_sharing'] )
