# -*- coding: utf-8 -*-
"""
This file is part of LionsTooth.

LionsTooth is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

LionsTooth is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with LionsTooth.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import pyqtSignal, pyqtSlot

from design.ui_profile import Ui_profile
from widgets.stream import stream as stream_widget
from widgets.markdown_viewer import markdown_viewer
from widgets.aspects_button import aspects_button, aspect_item
from widgets.shared.author_name import author_name
from widgets.photo_browser import photo_browser

from widgets.shared.tags_box import tags_box

""" Img downloading / caching """
from classes.resolver import img_worker

import markdown
from classes.markdown_extention import diaspora_md
import diaspy

class profile_aspects_button(aspects_button):
	add_to_aspect = pyqtSignal(int) # aspect id
	remove_from_aspect = pyqtSignal(int, str) # aspect id, aspect name
	def __init__(self, aspects=None, default_options=[], override_checked=None, parent=None):
		aspects_button.__init__(self, aspects=aspects, default_options=default_options, override_checked=override_checked, parent=parent)
		self.tmp_item_to_remove = None

		self.aspects_updated.connect( self.update_state )

	def data_changed(self, topLeft, bottomRight):
		self.model().dataChanged.disconnect( self.data_changed )

		item = self.model().itemFromIndex( topLeft )
		if type(item) is aspect_item:
			if item.checkState() == QtCore.Qt.Unchecked:
				self.tmp_item_to_remove = item
				self.remove_from_aspect.emit( item.id, item.name )

			elif item.checkState() == QtCore.Qt.Checked:
				self.add_to_aspect.emit( item.id )

		self.model().dataChanged.connect( self.data_changed )
		self.update_state() # Change color

	def update_state(self):
		if self.checked_count:
			self.setStyleSheet("background-color: #539a2d;\n"
			"color: #000;")
		else:
			self.setStyleSheet("background-color: #000;\n"
			"color: #fff;")

class photo_browser_window(QtWidgets.QFrame):
	deleting = pyqtSignal()
	delete_photo_request = pyqtSignal(int) # photo_id
	def __init__(self, photo_data, res, own_profile=False, parent=None):
		QtWidgets.QFrame.__init__(self, parent=parent)

		self.setWindowTitle("Photo Browser")

		self.layout = QtWidgets.QVBoxLayout(self)

		self.scroll_area = QtWidgets.QScrollArea(self)
		self.scroll_area.setWidgetResizable(True)

		scrollAreaWidgetContents = QtWidgets.QWidget()
		verticalLayout = QtWidgets.QVBoxLayout( scrollAreaWidgetContents )

		# Photo browser
		self.photo_browser = photo_browser(photo_data, res, own_profile=own_profile, parent=scrollAreaWidgetContents)

		sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
		self.photo_browser.setSizePolicy(sizePolicy)

		self.photo_browser.delete_photo_request.connect( self.delete_photo_request )
		
		verticalLayout.addWidget(self.photo_browser)

		# Set scroll area
		self.scroll_area.setWidget( scrollAreaWidgetContents )
		self.layout.addWidget( self.scroll_area )

		# Close button
		self.close_button = QtWidgets.QPushButton()
		self.close_button.setText("Close")
		self.layout.insertWidget( 0, self.close_button )
		self.close_button.clicked.connect( self.hideEvent )

	def hideEvent(self, event):
		self.deleting.emit()
		self.deleteLater()

class profile(QtWidgets.QFrame):
	aspects_updated = pyqtSignal()
	added_user_to_aspect = pyqtSignal(int, int, object) # contact_id, user_id, aspect_membership
	removed_user_from_aspect = pyqtSignal(int, int, int) # contact_id, user_id, aspect_id
	tag_clicked = pyqtSignal(str) # tag_name
	anchor_clicked = pyqtSignal(str) # str: link
	def __init__(self, res=None, parent=None):
		QtWidgets.QFrame.__init__(self, parent=parent)
		self.ui = Ui_profile()
		self.ui.setupUi(self)

		self._user = None		# diaspy.people.User needed for user data to fill in profile data
		self._contacts = None	# diaspy.people.Contacts needed to add and remove user from aspect(s)
		self._aspects = None	# diaspy.connection.getUserData()['aspects'] (list of your aspects) needed for aspects button, it contains aspect id and name

		self._res = res 		# resource for getting images
		self.max_width = 100
		self.own_profile = False

		""" aspect button """
		self.default_options = []
		self.ui.aspects_button = profile_aspects_button()
		self.ui.horizontalLayout_2.addWidget(self.ui.aspects_button, 0, QtCore.Qt.AlignTop) # TODO proper name for horizontalLayout_2

		""" bio """
		self.ui.bio = markdown_viewer()
		self.bio_html = None
		self.ui.bio_layout.addWidget( self.ui.bio )
		self.ui.bio.anchor_clicked.connect( self.anchor_clicked )

		""" tag box """
		self.tags_box = tags_box(max_width=self.max_width) # TODO max_with is probably None
		self.tags_box.tag_clicked.connect( self.tag_clicked )
		self.ui.horizontalLayout_2.insertWidget(0, self.tags_box )

		""" owner name """
		self.ui.name = author_name()
		self.ui.name.font = QtGui.QFont("FreeSans", 12)
		self.ui.name.setFont(self.ui.name.font)
		self.ui.name_layout.insertWidget(0, self.ui.name )

		""" Aspect button hooks """
		self.ui.aspects_button.remove_from_aspect.connect( self.remove_from_aspect )
		self.ui.aspects_button.add_to_aspect.connect( self.add_to_aspect )

		""" Photo browser and button hook """
		self.photo_browser_window = None
		self.ui.btn_photos.clicked.connect( self.open_photo_browser )

	def reset(self):
		# Name and diaspora_id
		self.ui.name.set_text("")
		self.ui.diaspora_id.setText("")

		# Avatar
		self.ui.avatar.hide()

		# Aspects button
		self.ui.aspects_button.hide()

		# Sharing status
		self.ui.sharing_message.setText("-")
		self.ui.sharing_message.setStyleSheet("color: #fff;")
		self.ui.sharing_message.setToolTip("Not sharing")

		# Bio
		self.ui.bio_box.hide()	# location, gender, birthday
		self.ui.bio.hide()		# markdown

		# Tags box
		self.tags_box.clear_layout()

		# Photo's button
		self.ui.btn_photos.hide()

	def set_res(self, res):
		self._res = res
		self.ui.bio.set_res( res )

	def set_max_width(self, max_width):
		self.max_width = max_width
		self.tags_box.set_max_width( max_width )
		self.ui.bio.resize_content(max_width=max_width, collapse=False)
		self.adjustSize()

	def delete_photo(self, photo_id):
		self._user.deletePhoto(photo_id)

		# let photo browser window know it's deleted if it still exists.
		if self.photo_browser_window:
			self.photo_browser_window.photo_browser.photo_deleted(photo_id)

	def open_photo_browser(self): # TODO subclass photo_browser
		if self._user.photos:
			if not self.photo_browser_window:
				self.photo_browser_window = photo_browser_window(self._user.photos, self._res, own_profile=self.own_profile)
				self.photo_browser_window.deleting.connect( self.close_photo_browser )
				self.photo_browser_window.move(0,0)
				self.photo_browser_window.showFullScreen()
				self.photo_browser_window.delete_photo_request.connect( self.delete_photo )

	def close_photo_browser(self):
		self.photo_browser_window = None

	def confirm(self, title, quesion): # TODO have this in more files so create own file for it
		answer = QtWidgets.QMessageBox.question(self, title, quesion, QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.No)
		if answer == QtWidgets.QMessageBox.Yes:
			return True
		return False

	def add_to_aspect(self, aspect_id): # TODO threads
		if self._contacts:
			try:# TODO
				new_aspect_membership = self._contacts.add(self._user.id(), [aspect_id]) # add excepts list for aspects_ids
				self._user.fetchguid(fetch_stream=False) # need to fetch page again to update aspectmembership ids TODO this is not needed 'cause new membership gets returned and should get updated by Diaspy
				#self.aspects_updated.emit()
				self.added_user_to_aspect.emit(self._user.data['contact']['id'], self._user.data['contact']['person_id'], new_aspect_membership)
			except diaspy.errors.AspectError as err:
				print("diaspy.errors.AspectError: {}".format(err))
				# TODO give a message

	def remove_from_aspect(self, aspect_id, aspect_name=None): # TODO thread
		if self._contacts:
			if self.confirm("Question", "Are you sure you want to remove {} from {}".format(self._user.data['name'], aspect_name)):
				try: # TODO
					self._contacts.remove(self._user, [aspect_id])
					self.removed_user_from_aspect.emit(self._user.data['contact']['id'], self._user.data['contact']['person_id'], aspect_id)
				except diaspy.errors.AspectError as err:
					print("diaspy.errors.AspectError: {}".format(err))
			else:
				self.ui.aspects_button.tmp_item_to_remove.set_checked(True)
				self.ui.aspects_button.tmp_item_to_remove = None

	def set_contacts(self, contacts):
		if contacts:
			self._contacts = contacts

	def set_aspects(self, aspects):
		if aspects:
			self._aspects = aspects

	#def update_bio(self, markdown):
	#	self.ui.bio.set_markdown( markdown )
	#	# TODO resize

	def update_avatar(self, path):
		pixmap = QtGui.QPixmap(path)
		if pixmap:
			self.ui.avatar.setPixmap(pixmap)
			if self.ui.avatar.isHidden(): self.ui.avatar.show()

	def set_user(self, user):
		"""user expects diaspy.people.User object
		"""
		if user:
			self._user = user
			self.update_data()

	def update_aspects_button(self):
		if not self._user: return # nothing to update
		if self._user.data['contact']:
			ids = [self._user.data['contact']['aspect_memberships'][i]['aspect']['id'] for i in range(len(self._user.data['contact']['aspect_memberships']))]
		else:
			ids = []
		aspects = []
		for a in self._aspects:
			selected = False
			if a['id'] in ids: selected = True
			aspects.append( { 
								'id'		: a['id'],
								'name'		: a['name'],
								'selected'	: selected
							}
			)
		self.ui.aspects_button.reset()
		self.ui.aspects_button.set_aspects( aspects )
		self.ui.aspects_button.update_aspects()

	def update_data(self):
		# Set username
		if self._user.data['name']:
			self.ui.name.set_text("{}".format( self._user.data['name'] ))

		# Diaspora ID
		if 'diaspora_id' in self._user.data:
			self.ui.diaspora_id.setText("<b>{}</b>".format( self._user.data['diaspora_id'] ))
		elif 'handle' in self._user.data:
			self.ui.diaspora_id.setText("<b>{}</b>".format( self._user.data['handle'] ))
		else:
			print("people.py update_data() key error, no diaspora_id or handle. userdata: {}".format(self._user.data))
			return

		if not self._user.data['is_own_profile']:
			# Aspect memberships dropdown
			self.ui.aspects_button.show()
			self.update_aspects_button()
			self.own_profile = False
		else:
			self.own_profile = True
			self.ui.aspects_button.hide()

		# Sharing_message status
		try:
			if self._user.data['relationship'] == "mutual":
				self.ui.sharing_message.setText("✓")
				self.ui.sharing_message.setToolTip("{} shares with you.".format(self._user.data['name']))
				self.ui.sharing_message.setStyleSheet("color: #1a8830;")
		except KeyError:
			print("KeyError in sharing_message status")

		# Photos button
		if self._user.photos:
			count = len(self._user.photos)
			self.ui.btn_photos.setText("{} Photo's".format(count))
			self.ui.btn_photos.show()

		if 'profile' in self._user.data:
			# Tags
			try:
				for tag in self._user.data['profile']['tags']:
					self.tags_box.add_tag(tag)
			except KeyError:
				print("KeyError in Tags")

			# Bio text
			visable_in_bio = 0
			if 'bio' in self._user.data['profile'] and self._user.data['profile']["bio"]:
				self.ui.bio.set_markdown( self._user.data['profile']["bio"] )
				if self.ui.bio.isHidden():
					self.ui.bio.show()
				self.ui.bio.resize_content(max_width=self.max_width, collapse=False)
				visable_in_bio +=1
			else:
				self.ui.bio.hide()

			# Birthday
			if 'birthday' in self._user.data['profile']:
				self.ui.birthday.setText(self._user.data['profile']["birthday"])
				if self.ui.birthday.isHidden():
					self.ui.birthday.show()
					self.ui.lbl_birthday.show()
				visable_in_bio +=1
			else:
				self.ui.birthday.hide()
				self.ui.lbl_birthday.hide()

			# Gender
			if 'gender' in self._user.data['profile']:
				self.ui.gender.setText(self._user.data['profile']["gender"])
				if self.ui.gender.isHidden():
					self.ui.gender.show()
					self.ui.lbl_gender.show()
				visable_in_bio +=1
			else:
				self.ui.gender.hide()
				self.ui.lbl_gender.hide()

			# Location
			if 'location' in self._user.data['profile']:
				self.ui.location.setText(self._user.data['profile']["location"])
				if self.ui.location.isHidden():
					self.ui.location.show()
					self.ui.lbl_location.show()
				visable_in_bio +=1
			else:
				self.ui.location.hide()
				self.ui.lbl_location.hide()

			""" hide bio groupbox if none labels are set 
			and show if it does have labels and is hidden """
			if visable_in_bio == 0:
				self.ui.bio_box.hide()
			elif self.ui.bio_box.isHidden():
				self.ui.bio_box.show()

#https://diaspora.github.io/api-documentation/routes/users.html
class people(stream_widget):
	def __init__(self, thread_pool, parent=None):
		stream_widget.__init__(self, parent=parent)
		self.name = "people"
		self.thread_pool = thread_pool

		self.profile = profile()
		self.profile.anchor_clicked.connect( self.anchor_clicked )

		self.ui.stream_layout.insertWidget(0, self.profile, 0, QtCore.Qt.AlignTop )

		self.ui.stream_layout.post_widgets = {}

		self.post_offset = 1

		self.res_set.connect( self.set_res_on_profile )
		self.max_width_set.connect( self.set_profile_max_width ) # needed for tags_box and markdown_viewer

	def set_res_on_profile(self):
		self.profile.set_res(self.res)

	def set_profile_max_width(self, max_width): # or self.max_width
		self.profile.set_max_width(max_width)
