"""
This file is part of LionsTooth.

LionsTooth is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

LionsTooth is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with LionsTooth.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import pyqtSignal

class aspect_item(QtGui.QStandardItem):
	def __init__(self, aspect_data, override_checked=None):
		QtGui.QStandardItem.__init__(self, aspect_data['name'])
		self.id = aspect_data['id']
		self.name = aspect_data['name']

		self.setFlags(QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsEnabled)
		if override_checked != None:
			self.set_checked(override_checked)
		else:
			self.set_checked(aspect_data['selected'])

	def set_checked(self, bool):
		if bool:
			self.setData(QtCore.Qt.Checked, QtCore.Qt.CheckStateRole)
		else:
			self.setData(QtCore.Qt.Unchecked, QtCore.Qt.CheckStateRole)

class aspects_button(QtWidgets.QComboBox):
	aspects_updated = pyqtSignal()
	def __init__(self, aspects=None, default_options=None, override_checked=None, parent=None):
		QtWidgets.QComboBox.__init__(self, parent=parent)
		self.aspects = aspects

		""" default """
		self.previous_text = "None"
		self.checked_count = 0
		self.override_checked = override_checked # override checked, set to None to do nothing, set to True or False to override checked state as default
		self.default_options = [
			{
				'id'		: 'public',
				'name'		: 'Public',
				'selected'	: False
			},{
				'id'		: 'all_aspects',
				'name'		: 'All aspects',
				'selected'	: True # TODO get default selected from settings (/user/edit)
			},
		]
		if default_options != None: # it may be empty, but not None for override
			self.default_options = default_options
		self.default_options_len = len(self.default_options)
		
		if self.default_options_len:
			model = self.default_items()
		else:
			model = QtGui.QStandardItemModel()

		view = QtWidgets.QColumnView()
		self.setView(view)
		
		sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
		self.setSizePolicy(sizePolicy)
		
		self.setModel(model)
		
		""" hooks """
		self.view().pressed.connect( self.toggle_item )
		self.model().dataChanged.connect( self.data_changed )
		
		""" separator """
		if default_options:
			self.insertSeparator( self.default_options_len )
			self.default_options_len += 1 # +1 for the separator

		""" on init """
		if aspects:
			self.update_aspects()

	def default_items(self):
		model = QtGui.QStandardItemModel()
		for i in range(self.default_options_len):
			item = aspect_item( self.default_options[i] )
			model.insertRow(i, item)
		return model

	def toggle_item(self, model_index):
		item = self.model().itemFromIndex( model_index )
		if item.checkState() == QtCore.Qt.Checked:
			item.set_checked(False)
		else:
			item.set_checked(True)

	""" title
	copied most of this from:
	https://stackoverflow.com/questions/47575880/qcombobox-set-title-text-regardless-of-items"""
	def unchecked_indexes(self):
		return self.model().match(self.model().index(0, 0), QtCore.Qt.CheckStateRole, QtCore.Qt.Unchecked, -1, QtCore.Qt.MatchRecursive)

	def checked_indexes(self):
		return self.model().match(self.model().index(0, 0), QtCore.Qt.CheckStateRole, QtCore.Qt.Checked, -1, QtCore.Qt.MatchRecursive)

	def paintEvent(self, event):
		painter = QtWidgets.QStylePainter(self)

		opt = QtWidgets.QStyleOptionComboBox()
		self.initStyleOption(opt)

		u_index = self.unchecked_indexes()
		c_index = self.checked_indexes()

		uncheck_count = len(u_index)
		checked_count = len(c_index)
		
		self.checked_count = checked_count

		if not checked_count: # none checked, this should not happen except on profile page
			opt.currentText = "Not aspect"

		elif not uncheck_count - self.default_options_len: # all checked, minus self.default_options_len
			opt.currentText = "All aspects"

		elif checked_count == 1: # one checked
			try:
				opt.currentText = self.model().itemFromIndex( c_index[0] ).name
			except AttributeError:
				pass

		else: # multiple checked
			opt.currentText = "In {} aspects".format(checked_count)

		painter.drawComplexControl(QtWidgets.QStyle.CC_ComboBox, opt)

		#draw the icon and text
		painter.drawControl(QtWidgets.QStyle.CE_ComboBoxLabel, opt)
			
		if self.previous_text != opt.currentText:
			self.previous_text = opt.currentText
			self.aspects_updated.emit()
		
	"""slot for dataChanged
	http://doc.qt.io/archives/qt-4.8/qabstractitemmodel.html#dataChanged"""
	def data_changed(self, topLeft, bottomRight):
		self.model().dataChanged.disconnect( self.data_changed )

		item = self.model().itemFromIndex( topLeft )
		if type(item) is aspect_item:
			if item.id == 'public':
				self.deselect(but='public')
			elif item.id == 'all_aspects':
				self.deselect(but='all_aspects')
				self.select_all_aspects()
			else:
				u_index = self.unchecked_indexes()
				minus = self.default_options_len
				if item.checkState() == QtCore.Qt.Unchecked: minus -=1 # else you can't deselect aspects
				if not len(u_index) - minus: # all aspects selected
					self.select_all_aspects()
				elif self.default_options:
					self.deselect_default()

		if not len(self.checked_indexes()):
			index = self.model().index(0,0) # TODO set default index
			self.model().setData(index, QtCore.Qt.Checked, QtCore.Qt.CheckStateRole)

		self.model().dataChanged.connect( self.data_changed )

	def selected_aspects(self): # returns a list with selected aspect id's
		ids = []
		for index in self.checked_indexes():
			item = self.model().itemFromIndex( index )
			ids.append( item.id )
		if 'all_aspects' in ids: # the ids will be there, it's enough
			ids.remove('all_aspects')
		return ids

	def select_all_aspects(self):
		for unchecked in self.unchecked_indexes():
			unchecked_item = self.model().itemFromIndex( unchecked )
			if unchecked_item.id != 'public':
				unchecked_item.set_checked(True)

	def deselect_default(self): # TODO this can be done more efficient
		for checked in self.checked_indexes():
			checked_item = self.model().itemFromIndex( checked )
			print(checked_item)
			if checked_item.id in ['public', 'all_aspects']:
				checked_item.set_checked(False)

	def deselect(self, but=None): # deselect everything but: public or all_aspects
		for checked in self.checked_indexes():
			checked_item = self.model().itemFromIndex( checked )
			if but and checked_item.id != but: # uncheck everything except but
				checked_item.set_checked(False)

	def set_aspects(self, aspects):
		self.aspects = aspects
	
	def refresh(self):
		self.reset()
		self.update_aspects()
	
	def update_aspects(self):
		#print("UPDATEING ASPECTS")#DEBUG
		if self.aspects:
			#print(self.aspects)#DEBUG
			model = self.model()
			for i in range(len(self.aspects)):
				index = self.default_options_len + i # +1 for the separator
				model.insertRow(index, aspect_item( self.aspects[i], override_checked=self.override_checked ))
			self.setModel(model)

	def reset(self):
		print("RESET ASPECTS")#DEBUG
		self.model().dataChanged.disconnect( self.data_changed )
		self.setModel( self.default_items() )
		self.model().dataChanged.connect( self.data_changed )
