# -*- coding: utf-8 -*-
"""
This file is part of LionsTooth.

LionsTooth is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

LionsTooth is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with LionsTooth.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import pyqtSignal

from design.ui_comment import Ui_comment
from design.ui_comments import Ui_comments

from widgets.shared.author_name import author_name
from widgets.shared.html_content import html_content

from classes.resolver import img_worker

import markdown
from classes.markdown_extention import diaspora_md
from widgets.markdown_viewer import markdown_viewer, more_less_button

from classes.formatter import tago, f_timestamp

"""
[
	{
		'id': 666, 
		'guid': '99e4d5d08fa20134a12600155da01107', 
		'text': 'Very nice comment', 
		'author': {
			'id': 123, 
			'guid': '99e4d5d08fa20134a12600155da01107', 
			'name': 'Test name', 
			'diaspora_id': 'test@test.test', 
			'avatar': {'small': '', 'medium': '', 'large': ''}
		}, 
		'created_at': '2018-03-14T10:42:00.000Z', 
		'mentioned_people': []
	}
]

"""
class comment(QtWidgets.QFrame):
	updated_content = pyqtSignal()
	author_name_clicked = pyqtSignal(str, str) # guid, diaspora_id
	anchor_clicked = pyqtSignal(str) # str: link
	delete_me = pyqtSignal(int) # id
	def __init__(self, comment, res, own_comment=False, parent=None):
		QtWidgets.QFrame.__init__(self, parent=parent)
		self.ui = Ui_comment()
		self.ui.setupUi(self)
		
		self.comment = comment
		self.own_comment = own_comment
		
		""" author name """
		self.ui.author_name = author_name( guid=self.comment.author(key='guid'), diaspora_id=self.comment.author(key='diaspora_id') )
		self.ui.author_name.set_text( self.comment.author(key='name') )
		self.ui.author_name.clicked.connect( self.author_name_clicked )
		
		""" time ago """
		self.update_tago()
		self.ui.tago.setToolTip( f_timestamp( self.comment.when() ) )

		""" delete button """
		if not self.own_comment:
			self.ui.btn_delete.hide()
		self.ui.btn_delete.clicked.connect( self.delete )

		""" avatar thread """
		worker = img_worker(res, url=self.comment.author(key='avatar')['small'])
		worker.signals.finished.connect(self.update_avatar)
		res.thread_pool.start(worker)
		
		""" comment text """
		self.ui.markdown_viewer = markdown_viewer(res=res)
		self.ui.markdown_viewer.set_markdown( str(self.comment) )
		self.ui.markdown_viewer.anchor_clicked.connect( self.anchor_clicked )
		
		""" add widgets """
		self.ui.content.insertWidget(0, self.ui.author_name)
		self.ui.content.insertWidget(1, self.ui.markdown_viewer)
		
		""" more / less button init """
		self.more_less_button = more_less_button(self)
		self.ui.content.addWidget( self.more_less_button )
		self.more_less_button.hide()
		
		self.ui.markdown_viewer.collapse_state.connect( self.more_less_button_state )
		
		""" resize hook """
		self.ui.markdown_viewer.content_updated.connect( self.updated_content )
	
	def select(self):
		self.setStyleSheet("QFrame#comment {\n"
		"	border-bottom: 2px solid #2f3d28;\n"
		"	border-left: 10px solid #4f5f4a;\n"
		"}")
	
	def delete(self):
		self.delete_me.emit( self.comment.id )
	
	def update_markdown_viewer(self, markdown_text):
		self.ui.markdown_viewer.set_markdown( markdown_text )
		#self.updated_content.emit()
	
	def update_avatar(self, path):
		pixmap = QtGui.QPixmap(path)
		if pixmap:
			self.ui.avatar.setPixmap(pixmap)
			
	def update_tago(self, created_at=None):
		if not created_at: created_at = self.comment.when()
		self.ui.tago.setText( tago(created_at) )
		
	def more_less_button_state(self, bool):
		if bool: # collapsed
			self.on_collapsed()
		else: # expanded
			self.on_expanded()
	
	def on_expanded(self):
		self.more_less_button.clicked.disconnect() # disconnect signal created by collapse()
		self.more_less_button.setText("less")
		self.more_less_button.clicked.connect( self.ui.markdown_viewer.collapse ) # connect collapse() to the more_less_button
	
	def on_collapsed(self):
		self.more_less_button.setText("more")
		self.more_less_button.show()
		try:
			self.more_less_button.clicked.disconnect() # disconnect signal created by expand()
		except TypeError:
			pass # not connected
		self.more_less_button.clicked.connect( self.ui.markdown_viewer.expand ) # connect expand() to the more_less_button

class comments(QtWidgets.QFrame):
	updated_content = pyqtSignal()
	author_name_clicked = pyqtSignal(str, str) # guid, diaspora_id
	anchor_clicked = pyqtSignal(str) # str: link
	del_comment = pyqtSignal(int) # comment id
	def __init__(self, res, comments=None, count=0, own_guid=None, own_post=False, parent=None, show_all=False):
		QtWidgets.QFrame.__init__(self, parent=parent)
		self.ui = Ui_comments()
		self.ui.setupUi(self)
		self.res = res
		self.count = count
		self.max_index = self.count
		self.max_width = 0
		
		self.__own_guid = own_guid # str: guid
		self.__own_post = own_post # bool: if own post or not
		self._comments = comments # list with diaspy.models.Comment's
		self.id_to_guid = {}
		self.__show_count = 3
		self.__collapsed = False
		if comments and len(comments) > self.__show_count:
			self.__collapsed = True

		if comments:
			self.update_comments()
			if show_all and self.max_index > self.__show_count: self.show_more()
		else:
			self.hide()

	def update_tago(self):
		current_comments = self.current_comments()
		for comment_id in current_comments:
			current_comments[comment_id].update_tago()

	def is_collapsed(self):
		return self.__collapsed

	def get_id(self, comment_guid):
		if comment_guid in self.guid_to_id:
			return self.guid_to_id[comment_guid]
		return None

	def update_guid_to_id(self):
		if self._comments:
			guid_to_id = {}
			for comment in self._comments:
				guid_to_id.update( { comment.guid: comment.id } )
			self.guid_to_id = guid_to_id

	def delete_comment(self, id):
		self.del_comment.emit(id)
		
	def remove_comment_from_layout(self, id):
		current_comments = self.current_comments()
		current_comments[id].author_name_clicked.disconnect()
		current_comments[id].updated_content.disconnect()
		current_comments[id].delete_me.disconnect()
		current_comments[id].deleteLater()

	def set_comments(self, comments):
		if comments: self._comments = comments

	def update_comments(self):
		current_comments = self.current_comments()
		updated_comment_ids = self._comments.ids() # TODO AttributeError: 'list' object has no attribute 'ids'
		#print("current_comments : {}".format(current_comments)) # DEBUG
		#print("updated_comment_ids : {}".format(updated_comment_ids)) # DEBUG

		""" delete removed comments """
		for id in current_comments:
			if id not in updated_comment_ids:
				current_comments[id].author_name_clicked.disconnect()
				current_comments[id].updated_content.disconnect()
				current_comments[id].delete_me.disconnect()
				current_comments[id].anchor_clicked.disconnect()
				current_comments[id].deleteLater()

		# Update comment count
		comment_count = 0
		if self._comments: comment_count = len(self._comments)
		self.count = comment_count

		# expand / collapse
		if comment_count <= self.__show_count:
			self.ui.btn_more.hide()
		elif self.__collapsed:
			begin_index = self.max_index - self.__show_count
			self.ui.btn_more.clicked.connect(self.show_more)
			self.ui.btn_more.setText("Show {} more comments".format(self.count-self.__show_count))
			self.ui.btn_more.show()

		# if self is hidden and we have comments, then show
		if comment_count and self.isHidden(): self.show()

		# update and insert comments
		begin_index = 0
		if self.__collapsed: begin_index = comment_count - self.__show_count
		for i in range(begin_index, comment_count):
			if self._comments[i].id in current_comments: # update, TODO users can't edit comments so this is not needed?
				current_comments[self._comments[i].id].update_markdown_viewer( str(self._comments[i]) )
			else: # insert
				own_comment=False
				if self.__own_post or self._comments[i].author(key='guid') == self.__own_guid:
					own_comment=True
				c = comment(self._comments[i], self.res, own_comment=own_comment, parent=self)

				c.author_name_clicked.connect( self.author_name_clicked )
				c.updated_content.connect( self.updated_content ) # if comment signals updated_content then this class does
				c.delete_me.connect( self.delete_comment )
				c.anchor_clicked.connect( self.anchor_clicked )

				if self.max_width:
					c.ui.markdown_viewer.resize_content( max_width=self.max_width )

				current_comments.update( { self._comments[i].id: c } ) # add to current_comments
				self.ui.comments_layout.insertWidget(i, c)

	def current_comments(self):
		""" Returns a dict with as key the comment id and as value the 
		comment widget wich is in the comments_layout"""
		current_comments = {}
		for i in range(self.ui.comments_layout.count()):
			item = self.ui.comments_layout.itemAt(i)
			if not item:
				continue
			w = item.widget()
			if w:
				if type(w) is comment:
					current_comments.update( { w.comment.id: w } )
		return current_comments

	def resize_comments(self, max_width=None): # also sets self.max_width if max_width
		if not max_width:
			max_width = self.max_width
		else:
			margins = 35+6+6+6+3+3 # avatar, 3x layout-spacing, layout-left, layout-right
			max_width -= margins
			self.max_width = max_width # set own max_width, used as reference for show_more()
		for i in range(self.ui.comments_layout.count()):
			item = self.ui.comments_layout.itemAt(i)
			if not item:
				continue
			w = item.widget()
			if w:
				if type(w) is comment:
					w.ui.markdown_viewer.resize_content( max_width=max_width )

	def author_name_clicked_func(self, guid, diaspora_id):
		self.author_name_clicked.emit(  )

	def show_more(self):
		self.ui.btn_more.setText("Hide {} comments".format(self.count-self.__show_count))
		self.ui.btn_more.clicked.disconnect(self.show_more)
		self.ui.btn_more.clicked.connect(self.show_less)
		count = 0
		for i in range(self.count-self.__show_count):
			own_comment=False
			if self.__own_post or self._comments[i].author(key='guid') == self.__own_guid: own_comment=True
			c = comment(self._comments[i], self.res, own_comment=own_comment, parent=self)
			c.updated_content.connect( self.updated_content ) # if comment signals updated_content then this class does
			self.ui.comments_layout.insertWidget(1+count, c)
			if self.max_width:
				c.ui.markdown_viewer.resize_content( max_width=self.max_width )
			count +=1
		self.__collapsed = False
			
	def show_less(self): # TODO: hide them, now everytime the resolver gets called when viewing them again etc.. or leave it 'cause it saves resources..
		self.ui.btn_more.setText("Show {} more comments".format(self.count-self.__show_count))
		self.ui.btn_more.clicked.disconnect(self.show_less)
		self.ui.btn_more.clicked.connect(self.show_more)
		while self.ui.comments_layout.count()-self.__show_count > 1:
			item = self.ui.comments_layout.takeAt(1)# at 1 'cause we want to leave the button
			if not item:
				continue
			w = item.widget()
			if w:
				if type(w) is comment:
					w.deleteLater()
		self.__collapsed = True
			

