"""
This file is part of LionsTooth.

LionsTooth is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

LionsTooth is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with LionsTooth.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import pyqtSignal

from design.ui_markdown_viewer import Ui_markdown_viewer

from widgets.shared.html_content import html_content
from widgets.shared.loading_icon import loading_icon

from classes.markdown_obj_parser import markdown_obj_parser

from classes.markdown_extention import diaspora_md
import markdown

from classes.resolver import img_worker

class more_less_button(QtWidgets.QPushButton):
	def __init__(self, parent=None):
		QtWidgets.QPushButton.__init__(self, parent=parent)

		self.setMaximumHeight(20)
		self.setObjectName("more_less_button")
		self.setAutoFillBackground(True)
		self.setStyleSheet("QPushButton#more_less_button:hover {\n"
"	background-color: qlineargradient(spread:pad, x1:0, y1:0, x2:0, y2:1, stop:0 rgba(50, 50, 50, 0), stop:1 rgba(50, 50, 50, 255));\n"
"	border-style: inset;\n"
"}\n"
"QPushButton#more_less_button {\n"
"	color: rgb(255, 255, 255);\n"
"	background-color: qlineargradient(spread:pad, x1:0, y1:0, x2:0, y2:1, stop:0 rgba(0, 0, 0, 0), stop:1 rgba(0, 0, 0, 255));\n"
"	border-style: inset;\n"
"}")
		self.setFlat(True)
		self.setAutoFillBackground(True)

class markdown_viewer(QtWidgets.QFrame):
	collapse_state = pyqtSignal(bool) # True if collapsed, False if expanded
	content_updated = pyqtSignal()
	anchor_clicked = pyqtSignal(str)
	def __init__(self, res=None, parent=None):
		QtWidgets.QFrame.__init__(self, parent=parent)
		self.ui = Ui_markdown_viewer()
		self.ui.setupUi(self)

		self.res = res
		self.parser = markdown_obj_parser()
		self.md = markdown.Markdown(extensions=[diaspora_md(), 'markdown.extensions.nl2br', 'markdown.extensions.tables'])
		self.collapse_height = 200
		self.collapsed = False
		self.max_width = None

		self.images={}

		""" DEBUG """
		#self.setStyleSheet("background-color: #FF0000;")

	def set_res(self, res):
		self.res = res

	def clear_markdown(self):
		while self.ui.content.count():
			item = self.ui.content.takeAt(0)
			if not item:
				continue
			w = item.widget()
			if w:
				widget_type = type(w)
				if widget_type in [html_content, image]:
					w.deleteLater()

	def set_markdown(self, markdown):
		self.clear_markdown() # clear
		objects = self.parser.parse( markdown )
		for obj in objects:
			if obj[0] == 'code_block':
				self.add_code_block(obj[1])
			elif obj[0] == 'text':
				self.add_html(obj[1])
			elif obj[0] == 'image':# 0:obj_name, 1:url, 2:alt_text, 3:title
				self.images.update( { obj[1] : image(alt=obj[2], title=obj[3], parent=self) } )
				if self.res:
					worker = img_worker(self.res, url=obj[1])
					worker.signals.finished.connect( self.images[obj[1]].update_image )
					self.res.thread_pool.start(worker)
				else: print("no res in markdown_viewer") # DEBUG
				self.images[obj[1]].image_updated.connect( self.content_updated )
				self.ui.content.addWidget( self.images[obj[1]] )
				self.images[obj[1]].show()
			elif obj[0] == 'image_link':
				self.images.update( { obj[1] : image(alt=obj[2], title=obj[3], link=obj[5], parent=self) } )
				if self.res:
					worker = img_worker(self.res, url=obj[1])
					worker.signals.finished.connect( self.images[obj[1]].update_image )
					self.res.thread_pool.start(worker)
				self.images[obj[1]].image_updated.connect( self.content_updated )
				self.ui.content.addWidget( self.images[obj[1]] )
		self.content_updated.emit()

	def resize_content(self, max_width=None, collapse=True):
		if not max_width:
			max_width = self.max_width
		else:
			self.max_width = max_width
		for i in range(self.ui.content.count()):
			item = self.ui.content.itemAt(i)
			if not item:
				continue
			w = item.widget()

			if w:
				if type(w) is html_content:
					w.resize_content( max_width=max_width )

				elif type(w) is image:
					w.rescale( max_width=max_width )
		if collapse:
			self.collapse()

	def collapse(self):
		self.updateGeometry()
		content_height = self.sizeHint().height()
		sub = content_height - self.collapse_height
		if sub > 100: # 100 margin; only if content size if 100 pixels larger then collapse_height
			self.collapsed = True
			sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
			self.setSizePolicy(sizePolicy)
			self.setMaximumHeight( self.collapse_height )
			self.collapse_state.emit(True)
		else:
			self.collapsed = False
			self.setMinimumHeight( 0 )
			self.setMaximumHeight( 16777215 )

	def expand(self):
		""" expand content (html_content, photo_browser, poll) if collapsed """
		if self.collapsed:
			self.collapsed = False

			sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
			self.setSizePolicy(sizePolicy)
			self.setMinimumHeight( 0 )
			self.setMaximumHeight( 16777215 )

			self.collapse_state.emit(False)

	def stop_animations(self):
		for img_path in self.images:
			if self.images[img_path].is_animation():
				self.images[img_path].stop_animation()

	def start_animations(self):
		for img_path in self.images:
			if self.images[img_path].is_animation():
				self.images[img_path].start_animation()

	def add_html(self, text):
		content = html_content()
		content.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
		content.anchor_clicked.connect( self.anchor_clicked )
		html =  self.md.convert( text )
		self.ui.content.addWidget( content )
		content.update_html_content( html )

	def add_code_block(self, text):
		new_text = ""
		for line in text.splitlines(True):
			new_text += "	{}".format(line) # python-markdown uses 4 spaces for code
		content = html_content()
		html =  self.md.convert( new_text )
		self.ui.content.addWidget( content )
		content.update_html_content( html )

class quote(QtWidgets.QLabel):
	def __init__(self, text=None, parent=None):
		QtWidgets.QLabel.__init__(self, parent=parent)
		self.setIndent(42)
		if text:
			self.setText(text)

from PIL import Image	
class image(QtWidgets.QLabel):
	image_updated = pyqtSignal()
	link_pressed = pyqtSignal(str) # str: the link
	def __init__(self, title=None, alt=None, img_path=None, link=None, parent=None):
		QtWidgets.QLabel.__init__(self, parent=parent)

		self.pixmap = None
		self.movie = None

		self.original_size = [0,0]
		self.title = title
		self.alt = alt
		self.link = link
		self.img_path = img_path

		if title and not link: self.setToolTip(self.title)

		self.setText( self.alt )

		self._is_animation = False
		self._not_image = True

		self.gif_play = QtWidgets.QLabel(self)
		self.gif_play.setAttribute( QtCore.Qt.WA_TranslucentBackground )
		self.gif_play_pixmap = QtGui.QPixmap(":/icons/icons/gif_play.svg")
		self.gif_play.setPixmap( self.gif_play_pixmap )
		self.gif_play.hide()

		if link:
			self.link_w = QtWidgets.QLabel(self)
			self.link_w.setStyleSheet("background-color: #3f5834;")
			self.link_w.setText( "<a href=\"{}\">{}</a>".format(link, link) )
			self.link_w.setToolTip( link )
			self.link_w.hide()

			self.setCursor( QtCore.Qt.PointingHandCursor )
			self.setToolTip( link ) # bypass title, I find knowing what I'm clicking more important.

		if img_path: self.update_image(img_path)

	def is_animation(self): return self._is_animation

	def enterEvent(self, event):
		if self.link:
			self.setStyleSheet("border: 1px solid #AA0000")
			self.link_w.show()

	def leaveEvent(self, event):
		if self.link:
			self.link_w.hide()
			self.setStyleSheet("border: 0px")

	def _show_play(self):
		if not self.movie: return
		movie_size = self.movie.scaledSize()
		play_size = self.gif_play.size()

		pos_x = (movie_size.width() / 2) - (play_size.width() / 2)
		pos_y = (movie_size.height() / 2) - (play_size.height() / 2)

		self.gif_play.move(pos_x, pos_y)
		self.gif_play.show()

	def stop_animation(self):
		if self.movie:
			self._show_play()
			self.movie.setPaused(True)

	def start_animation(self):
		if self.movie:
			self.gif_play.hide()
			self.movie.setPaused(False)

	def mousePressEvent(self, event):
		button = event.button()
		if button == 1: # left mouse button
			if self._is_animation and self.movie: # TODO check if _is_animation should be enough, for now it isn't.
				if self.movie.state() == 2: # QMovie::Running
					self.movie.setPaused(True)
					self._show_play()
				else:
					self.movie.setPaused(False)
					self.gif_play.hide()
			elif self.link: # Links cannot be pressed this way with a animation, but why do you want that.. hmm?
				self.link_pressed.emit(self.link)
				print(self.link) # DEBUG
		elif button == 2: # right mouse button
			print(self.img_path)

	def update_image(self, img_path):
		if img_path:
			try:
				img_test = Image.open(img_path)
				self.original_size = img_test.size
			except OSError:
				print("DEBUG markdown_viewer update_image(img_path) NOT A IMAGE: {}".format(img_path)) #DEBUG
				self._not_image = True
				return

			self.img_path = img_path
			self.clear()
			self._not_image = False # Pillow can open it, so it's a image.

			if img_test.format == "GIF" and 'loop' in img_test.info: # if there is no loop then it's not animated.
				self._is_animation = True
				self.movie = QtGui.QMovie( img_path ) # TODO if self.movie is None here, then try pixmap
				self.setMovie( self.movie )
				self.movie.start()
				self._show_play()
				self.movie.setPaused(True)
			else:
				self.pixmap = QtGui.QPixmap( img_path )
				self.setPixmap( self.pixmap )
				self.movie = None

			img_test.close() # close PIL img
			self.image_updated.emit()

	def rescale(self, max_width=None):
		""" if it's not an image """
		if self._not_image == True:
			self.setMaximumWidth( max_width ) # TODO wordwrap function?
			return

		""" if it is a image """
		if self.original_size[0] > max_width:
			self.setMaximumWidth( max_width )
			if self.pixmap:
				scaled_pixmap = self.pixmap.scaledToWidth( max_width )
				self.setPixmap( scaled_pixmap )

			elif self.movie:
				self.start_animation()
				if self.original_size[0] == 0: return
				size = QtCore.QSize( self.original_size[0], self.original_size[1] )
				size.scale( max_width, max_width, QtCore.Qt.KeepAspectRatio )# scale is a void

				self.movie.setScaledSize( size )
				self.setMovie( self.movie )
				self.stop_animation()
		else:
			self.setMaximumWidth( self.original_size[0] )
			if self.pixmap:
				self.setPixmap( self.pixmap )
			elif self.movie:
				self.start_animation()
				size = QtCore.QSize( self.original_size[0], self.original_size[1] )
				self.movie.setScaledSize( size )
				self.setMovie( self.movie )
				self.stop_animation()
