# -*- coding: utf-8 -*-
"""
This file is part of LionsTooth.

LionsTooth is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

LionsTooth is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with LionsTooth.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import pyqtSignal

from design.ui_poll import Ui_poll
from design.ui_poll_option import Ui_poll_option

from widgets.shared.wrappable_text import wrappable_text

class poll_option(QtWidgets.QWidget):
	radio_selected = pyqtSignal(int)
	def __init__(self, data, already_participated=False, vote_multiplier=0, parent=None):
		QtWidgets.QWidget.__init__(self, parent=parent)
		self.ui = Ui_poll_option()
		self.ui.setupUi(self)
		
		self.data = data
		self.vote_multiplier = vote_multiplier
		self.setObjectName( str(self.data['id']) )
		self.ui.option_text = wrappable_text(self)
		self.ui.option_layout.addWidget( self.ui.option_text )
		self.ui.option_text.update_text( self.data['answer'] )

		self.ui.option_radio.toggled.connect( self.selected )

		self.update_progress()

		if already_participated: self.participated(True)

	def update_data(self, option_data, already_participated=False, vote_multiplier=0):
		self.data = option_data
		if vote_multiplier: self.vote_multiplier = vote_multiplier
		if already_participated: self.participated(True)
		self.update_progress()

	def update_progress(self):
		percent = self.data['vote_count'] * self.vote_multiplier
		self.ui.option_progress.setProperty("value", percent)
		self.ui.option_progress.setFormat("%p% ({} votes)".format(self.data['vote_count']))
	
	def participated(self, bool=True):
		if bool:
			self.ui.option_radio.hide()
			self.show_result(True)
		else:
			self.ui.option_radio.show()
	
	def show_result(self, bool=True):
		if bool:
			self.ui.option_progress.show()
		else:
			self.ui.option_progress.hide()
	
	def selected(self):
		self.radio_selected.emit(self.data['id'])
		self.ui.option_text.set_bold(True)
		
	def deselect(self):
		self.ui.option_radio.setChecked(False)
		self.ui.option_text.set_bold(False)

	def select(self):
		self.ui.option_radio.setChecked(True)
		self.ui.option_text.set_bold(True)

class poll(QtWidgets.QFrame):
	resized = pyqtSignal()
	did_vote = pyqtSignal(int, int) # int: poll_id, int: poll_answer_id
	def __init__(self, data, already_participated=False, answer=None, parent=None):
		QtWidgets.QWidget.__init__(self, parent=parent)
		self.ui = Ui_poll()
		self.ui.setupUi(self)
		
		self.data = data # Poll data
		self.selected_id = answer
		self.already_participated = already_participated
		
		self.vote_multiplier = 0
		if self.data['participation_count']:
			self.vote_multiplier = 100 / self.data['participation_count']
		
		""" Questing text """
		self.ui.poll_question = wrappable_text(self)
		self.ui.poll_question.update_text( "{}".format(self.data['question']) )
		self.ui.poll_question.set_bold(True)
		
		self.ui.poll_layout.insertWidget(0, self.ui.poll_question)
		
		""" Add options """
		self.option = {}
		for option_data in self.data['poll_answers']:
			self.option.update( { option_data['id']: poll_option(option_data, already_participated=self.already_participated, vote_multiplier=self.vote_multiplier) } )
			self.option[ option_data['id'] ].radio_selected.connect( self.change_toggled )
			self.ui.poll_options.addWidget( self.option[ option_data['id'] ] )
		
		if self.already_participated:
			self.option[self.selected_id].select()
			self.ui.btn_vote.hide()
		else:
			self.ui.btn_vote.clicked.connect( self.vote )

	def update_data(self, data):
		self.data = data
		if self.data['participation_count']:
			self.vote_multiplier = 100 / self.data['participation_count']
		for option_id in self.option:
			option_data = {}
			for o in data['poll_answers']:
				if o['id'] == option_id:
					option_data = o
					break
			self.option[option_id].update_data(option_data, already_participated=self.already_participated, vote_multiplier=self.vote_multiplier)

	def vote(self):
		if self.selected_id and not self.already_participated:
			self.did_vote.emit(self.data['poll_id'], self.selected_id)	# Emit that has been voted, so vote can be send

	def voted(self):
		if self.selected_id and not self.already_participated:
			self.ui.btn_vote.hide()										# Hide vote button
			self.ui.btn_vote.clicked.disconnect( self.vote )			# Disconnect vote button from this function
			
			self.option[ self.selected_id ].data['vote_count'] += 1		# Add the vote count to the vote id
			self.data['participation_count'] += 1						# Add the vote count to participation_count so the multiplier can be recalculated
			self.vote_multiplier = 100 / self.data['participation_count']
			
			for id in self.option:
				self.option[id].vote_multiplier = self.vote_multiplier	# Set the new multiplier on each poll option
				self.option[id].update_progress()						# Update the progressbar with the new data
				self.option[id].participated()							# Hide radio buttons, show progressbar and make option where voted for bold

	def change_toggled(self, id):
		if self.selected_id:
			self.option[ self.selected_id ].deselect()
		self.selected_id = id
	
	def resize_text(self):
		self.ui.poll_question.fit_height()
		
		for i in range(self.ui.poll_options.count()):
			item = self.ui.poll_options.itemAt(i)
			if not item:
				continue
			widget = item.widget()
			if type(widget) is poll_option:
				widget.ui.option_text.fit_height()
