# -*- coding: utf-8 -*-
"""
This file is part of LionsTooth.

LionsTooth is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

LionsTooth is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with LionsTooth.  If not, see <http://www.gnu.org/licenses/>.
"""

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import pyqtSignal, pyqtSlot

from design.ui_post import Ui_post

from widgets.post_interaction_bar import post_interaction_bar
from widgets.post_controls import post_controls
from widgets.photo_browser import photo_browser
from widgets.open_graph import open_graph
from widgets.comments import comments
from widgets.poll import poll

from widgets.markdown_viewer import markdown_viewer, more_less_button
from widgets.markdown_editor import markdown_editor

from classes.handler import post_worker # threads

from widgets.shared.html_content import html_content
from widgets.shared.author_name import author_name

from classes.resolver import img_worker
from classes.markdown_extention import diaspora_md
from classes.formatter import tago, f_timestamp
from PIL import Image # to get img dimensions
import markdown
import re

from datetime import datetime

class post_widget(QtWidgets.QWidget):
	resized = pyqtSignal(str, str) # str: guid, str: page_name
	author_clicked = pyqtSignal(str, str) # str: guid, diaspora_id
	anchor_clicked = pyqtSignal(str) # str: link
	permalink_clicked = pyqtSignal(str) # str: guid

	def __init__(self, res, post=None, post_data=None, thread_pool=None, own_guid=None, own_avatar=None, page_name=None, from_reshare=False, max_width=None, show_all_comments=False, parent=None):
		QtWidgets.QWidget.__init__(self, parent=parent)
		self.ui = Ui_post()
		self.ui.setupUi(self)
		
		self.max_width = max_width

		self.post = None
		if post:
			self.post = post
			if not post_data: post_data = post._data

		self.res = res
		self.thread_pool = thread_pool
		self.from_reshare = from_reshare
		self.is_reshare = False
		self.ui_data = post_data
		self.collapse_height = 200
		self.collapsed = False
		self.poll = None
		self.has_markdown_content = False
		self.has_photo_browser = False
		self.page_name = page_name
		
		self.id = None
		if 'id' in post_data:
			self.id = post_data['id']
		if not post_data['guid']: # TODO fix this hack
			self.guid = post_data['id']
		else:
			self.guid = post_data['guid']
			
		self.setObjectName( str(self.guid) )
		
		# TODO update all post_data variables to self.post._data
		self.author_guid = post_data['author']['guid']
		self.author_name = post_data['author']['name']
		self.diaspora_id = post_data['author']['diaspora_id']
		
		self.comments = None
		self.own_guid = own_guid
		self.own_post = False
		self.own_avatar = own_avatar
		if self.author_guid == own_guid:
			self.own_post = True
		
		""" avatar thread """
		worker = img_worker(self.res, url=self.ui_data['author']['avatar']['small'])
		worker.signals.finished.connect(self.update_avatar)
		self.res.thread_pool.start(worker)

		self.markdown_content = post_data['text']

		""" self """
		sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
		sizePolicy.setHorizontalStretch(0)
		sizePolicy.setVerticalStretch(0)
		sizePolicy.setHeightForWidth(self.sizePolicy().hasHeightForWidth())
		self.setSizePolicy(sizePolicy)

		""" scope """
		self.public = False
		if post_data['public']:
			scope_str = "public"
			self.public = True
		else: scope_str = "private"
		if post_data['provider_display_name']: scope_str+= " via <b>{}</b>".format(post_data['provider_display_name'])

		""" change stylesheet if from reshare """
		if self.from_reshare:
			self.ui.frame.setStyleSheet("background-color: rgb(26, 37, 24);\n"
			"color: rgb(208, 221, 255);")
		else:
			self.ui.frame.setStyleSheet("background-color: rgb(21, 30, 20);\n"
			"color: rgb(208, 221, 255);")

		""" author name """
		self.ui.lbl_author = author_name(guid=self.author_guid, diaspora_id=self.diaspora_id)
		self.ui.lbl_author.set_text( self.author_name )
		self.ui.hl_topbar.insertWidget(0, self.ui.lbl_author)
		self.ui.lbl_author.clicked.connect( self.author_clicked_func ) # author name clicked

		""" post controls """
		self.is_subscribed = False
		if str(post_data['participation']).lower() == 'true':
			self.is_subscribed = True
		self.ui.post_controls = post_controls(own_post=self.own_post, is_subscribed=self.is_subscribed)
		self.ui.hl_topbar.addWidget( self.ui.post_controls )

		self.ui.post_controls.signals.subscribe.connect( self.subscribe )
		self.ui.post_controls.signals.mute.connect( self.mute )
		self.ui.post_controls.signals.hide.connect( self.hide_post )
		self.ui.post_controls.signals.delete.connect( self.delete )
		self.ui.post_controls.signals.report.connect( self.report )

		""" permalink """
		self.ui.btn_permalink.clicked.connect( self.permalink )

		""" time ago """
		self.created_at = post_data['created_at']
		self.ui.lbl_timeago.setToolTip( f_timestamp(self.created_at) )
		self.update_tago()
		
		"""
		TODO
		 * move this timer to main window and call functions from there
		 * only if mainwindow is active and post is visable through viewport?
		 * update comments timeago
		"""
		self.tago_timer = QtCore.QTimer()
		self.tago_timer.timeout.connect( self.update_tago )
		self.tago_timer.start(1000*60*5) # 5 minutes 

		self.reshare_post = None
		if post_data['post_type'].lower() == 'reshare':
			""" if reshare add another post inside this post """
			self.is_reshare = True
			if post_data['root']:
				self.reshare_post = post_widget(parent=self.ui.frame, post_data=post_data['root'], res=self.res, own_guid=self.own_guid, from_reshare=True)
				self.reshare_post.ui.lbl_author.clicked.connect( self.author_clicked_func ) # author name clicked
				self.reshare_post.anchor_clicked.connect( self.anchor_clicked ) # anchor clicked
				self.reshare_post.permalink_clicked.connect( self.permalink_clicked ) # permalink clicked
				self.ui.content_layout.addWidget(self.reshare_post)
			else:
				""" TODO
				if root is null, then post that is reshared is deleted.
				
				Diaspora sets text to:
				'	Originele bericht verwijderd door de auteur.'
				"""
				pass

		else: # markdown content
			self.ui.markdown_viewer = markdown_viewer(res=self.res, parent=self.ui.frame)
			self.ui.markdown_viewer.set_markdown( self.markdown_content )
			self.ui.markdown_viewer.collapse_state.connect( self.more_less_button_state )
			self.ui.markdown_viewer.content_updated.connect( self.resize_event )
			self.ui.markdown_viewer.anchor_clicked.connect( self.anchor_clicked )

			self.ui.content_layout.addWidget( self.ui.markdown_viewer )
			self.has_markdown_content = True

			""" Photo browser """
			if post_data['photos']:
				#self.photo_browser = photo_browser(post_data['photos'], self.res, parent=self.ui.content)
				self.photo_browser = photo_browser(post_data['photos'], self.res)
				self.photo_browser.resized.connect( self.resize_photo_browser )
				self.photo_browser.author_name_clicked.connect( self.author_clicked_func )
				self.ui.content_layout.insertWidget(0, self.photo_browser, 0, QtCore.Qt.AlignCenter)
				self.has_photo_browser = True

			""" Poll """
			if post_data['poll']:
				already_participated = False
				answer = None
				if 'poll_participation_answer_id' in post_data:
					if post_data['poll_participation_answer_id']:
						#if str(post_data['participation']).lower() == 'true': # apparently they don't use the value of participation anymore, or it's broken.
						already_participated = True
						answer = post_data['poll_participation_answer_id']
				self.poll = poll(post_data['poll'], already_participated=already_participated, answer=answer)
				self.poll.did_vote.connect( self.vote_poll )
				self.ui.content_layout.insertWidget(1, self.poll, 0, QtCore.Qt.AlignCenter)
				self.poll.resize_text()

			""" more / less button init """
			self.more_less_button = more_less_button(self)
			self.ui.content_layout.addWidget(self.more_less_button)
			self.more_less_button.hide()

			""" Open Graph """
			if post_data['open_graph_cache']:
				self.open_graph = open_graph(post_data['open_graph_cache'], self.res)
				self.open_graph.anchor_clicked.connect(self.anchor_clicked) # Anchor hook
				self.ui.content_layout.addWidget(self.open_graph)

		""" Interactions (Like, Share, Comment) """
		#if not from_reshare:# only show in parent post
		self.interaction_bar = post_interaction_bar( post_data['interactions'], own_guid=self.own_guid, own_post=self.own_post, public=self.public )
		self.interaction_bar.ui.lbl_scope.setText(scope_str)

		self.interaction_bar.signals.like.connect( self.like )
		self.interaction_bar.signals.reshare.connect( self.reshare )

		if not from_reshare:
			self.interaction_bar.ui.btn_comment.clicked.connect( self.toggle_comment_editor )
		else:# from_reshare doesn't have Post only data so disable buttons
			self.interaction_bar.ui.btn_comment.hide()
			self.interaction_bar.ui.btn_like.hide()
			self.interaction_bar.ui.btn_reshare.hide()

		self.ui.content_layout.addWidget(self.interaction_bar)

		""" comments """
		if not from_reshare:
			c = self.post.comments
			self.comments = comments(self.res, comments=c, parent=self.ui.frame, own_guid=self.own_guid, own_post=self.own_post, count=post_data['interactions']['comments_count'], show_all=show_all_comments)
			self.ui.content_layout.addWidget(self.comments)
			self.comments.updated_content.connect( self.resize_comments )
			self.comments.author_name_clicked.connect( self.author_clicked_func )
			self.comments.del_comment.connect( self.delete_comment )
			self.comments.anchor_clicked.connect( self.anchor_clicked )
						
			# markdown editor, this should be in comments widget, but for the layout thing it's better here for now.
			self.markdown_editor = markdown_editor(res=self.res, is_comment=True, max_width=self.max_width)
			
			self.markdown_editor.submit_request_comment.connect( self.comment )
			
			if self.own_avatar: # TODO if user updated avatar, it won't update here now.
				self.markdown_editor.set_avatar( path=self.own_avatar ) # set own avatar in markdown editor
			self.ui.content_layout.addWidget(self.markdown_editor)
			if not self.post.comments:
				self.markdown_editor.hide()

		""" DEBUG """
		self.ui.btn_refresh.clicked.connect( self.update )
		#if post_data['o_embed_cache']:
		#	print(post_data['o_embed_cache'])
		
		self.show()

	def update_data(self):
		# Interaction data
		self.interaction_bar.set_data( self.post._data['interactions'] )

		# Comments
		if self.post.comments:
			self.comments.set_comments( self.post.comments )
			self.comments.update_comments()

		# Poll
		if self.poll: self.poll.update_data(self.post._data['poll'])

		# Refresh images?
		# ..

	def update(self):
		if self.post:
			worker = post_worker('interaction', interaction='update', post=self.post)
			worker.signals.success.connect( self.update_data )
			#worker.signals.exception.connect(  ) # TODO print msg
			self.thread_pool.start(worker)

	def confirm(self, title, quesion):
		answer = QtWidgets.QMessageBox.question(self, title, quesion, QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.No)
		if answer == QtWidgets.QMessageBox.Yes:
			return True
		return False

	""" interaction slots """
	def toggle_comment_editor(self):
		if self.markdown_editor.isHidden():
			self.markdown_editor.show()
			self.markdown_editor.expand()
		else:
			self.markdown_editor.hide()

	def comment_successfull_added(self, comment): # comment: diaspy.models.Comment object
		self.comments.update_comments() # let the comments class insert the comment into the layout
		self.interaction_bar.commented() # add +1 to interaction_bar
		self.markdown_editor.published() # let the markdown editor know it's published

	def comment(self, markdown):
		worker = post_worker('comment', comment_markdown=markdown, post=self.post)
		worker.signals.new_comment.connect( self.comment_successfull_added )
		worker.signals.exception.connect( self.markdown_editor.not_published )
		self.thread_pool.start(worker)

	def deleted_comment(self, id):
		self.comments.remove_comment_from_layout(id) # Remove comment from layout
		self.interaction_bar.comment_deleted()# When succesfully deleted sub 1 from interaction_bar

	def delete_comment(self, id):
		if self.confirm("Question", "Are you sure you want to delete this comment?"):
			worker = post_worker('interaction', interaction='delete_comment', delete_comment_id=id, post=self.post)
			worker.signals.success.connect( lambda: self.deleted_comment(id) )
			#worker.signals.exception.connect( ) # TODO print message
			self.thread_pool.start(worker)

	def like(self, bool): # like or unlike
		if bool:
			worker = post_worker('interaction', interaction='like', post=self.post)
			worker.signals.success.connect( self.interaction_bar.liked )
		else:
			worker = post_worker('interaction', interaction='unlike', post=self.post)
			worker.signals.success.connect( self.interaction_bar.unliked )
		worker.signals.exception.connect( self.interaction_bar.not_liked )
		self.thread_pool.start(worker)

	def reshare(self): # reshare
		worker = post_worker('interaction', interaction='reshare', post=self.post)
		worker.signals.success.connect( self.interaction_bar.reshared )
		worker.signals.exception.connect( self.interaction_bar.not_reshared )
		self.thread_pool.start(worker)

	def vote_poll(self, poll_id, poll_answer_id):
		worker = post_worker('interaction', interaction='vote_poll', post=self.post, vote_answer_id=poll_answer_id)
		worker.signals.success.connect( self.poll.voted )
		#worker.signals.exception.connect(  ) # TODO again msg..
		self.thread_pool.start(worker)

	""" permalink """
	def permalink(self):
		self.permalink_clicked.emit(self.guid)

	""" post control functions """
	def report(self):
		pass

	def delete(self):
		worker = post_worker('interaction', interaction='delete', post=self.post)
		worker.signals.success.connect( self.hide )
		#worker.signals.exception.connect(  ) # TODO print msg that it failed remote but hide anyway for now.
		self.thread_pool.start(worker)

	def hide_post(self):
		worker = post_worker('interaction', interaction='hide', post=self.post)
		worker.signals.success.connect( self.hide )
		#worker.signals.exception.connect(  ) # TODO print msg that it failed remote but hide anyway for now.
		self.thread_pool.start(worker)

	def mute(self):
		worker = post_worker('interaction', interaction='mute', post=self.post)
		worker.signals.success.connect( self.hide )
		#worker.signals.exception.connect(  ) # TODO print msg that it failed remote but hide anyway for now.
		self.thread_pool.start(worker)

	def subscribe(self, bool):
		if bool:
			worker = post_worker('interaction', interaction='subscribe', post=self.post)
			worker.signals.success.connect( self.ui.post_controls.subscribed )
		else:
			worker = post_worker('interaction', interaction='unsubscribe', post=self.post)
			worker.signals.success.connect( self.ui.post_controls.unsubscribed )
		#worker.signals.exception.connect( self.interaction_bar.not_reshared ) # TODO agian print msg
		self.thread_pool.start(worker)

	""" post functions """
	def update_tago(self): # Update time ago
		timeago = tago(self.created_at)
		self.ui.lbl_timeago.setText(timeago)
		if self.comments:
			self.comments.update_tago()

	def resize_event(self):
		self.set_max_width(self.max_width)

	def set_max_width(self, max_width, layout=None):
		self.setMaximumWidth( max_width )
		self.max_width = max_width

		if not layout:
			layout = self.ui.content_layout

		if self.from_reshare:
			margin_width_content = 35+8+8+8+2 # avatar, 3x horizontalLayout_2 widget spacing, verticalLayout_2 left-margin
		else:
			margin_width_content = 50+8+8+8+2
		max_width -=margin_width_content

		for i in range(0, layout.count()):
			item = layout.itemAt(i)
			if not item:
				continue
			w = item.widget()

			if w:
				w_type = type(w)
				if w_type is markdown_viewer:
					w.resize_content(max_width=max_width) # markdown viewer will also collapse if needed

				elif w_type is markdown_editor:
					w.set_max_width(max_width=max_width)
					w.collapse()

				elif w_type is post_widget: # reshare
					w.set_max_width(max_width=max_width+35, layout=self.ui.content_layout)
					
				elif w_type is photo_browser:
					w.resize_photo(max_width=max_width)
					
				elif w_type is comments:
					w.resize_comments(max_width=max_width)

	def resize_photo_browser(self, max_width=None):
		if not max_width:
			max_width = self.max_width
		self.photo_browser.resize_photo(max_width=max_width)
	
	def resize_markdown_content(self, max_width=None):
		if not max_width:
			max_width = self.max_width
		self.ui.markdown_viewer.resize_content(max_width=max_width)
		
	def resize_comments(self, max_width=None):
		if not max_width:
			max_width = self.max_width-64
		self.comments.resize_comments(max_width=max_width)

	def author_clicked_func(self, guid, diaspora_id):
		self.author_clicked.emit(guid, diaspora_id)
	
	def more_less_button_state(self, bool):
		if bool: # collapsed
			self.on_collapsed()
		else: # expanded
			self.on_expanded()
	
	def on_expanded(self):
		self.more_less_button.clicked.disconnect() # disconnect signal created by collapse()
		self.more_less_button.setText("less")
		self.more_less_button.clicked.connect( self.ui.markdown_viewer.collapse ) # connect collapse() to the more_less_button
	
	def on_collapsed(self):
		self.more_less_button.setText("more")
		self.more_less_button.show()
		try:
			self.more_less_button.clicked.disconnect() # disconnect signal created by expand()
		except TypeError:
			pass # not connected
		self.more_less_button.clicked.connect( self.ui.markdown_viewer.expand ) # connect expand() to the more_less_button

	def update_avatar(self, path):
		""" avatar """
		if path:
			pixmap = QtGui.QPixmap(path)
			if pixmap:
				if self.from_reshare:
					pixmap.scaledToWidth(35)
					pixmap.scaledToHeight(35)
					self.ui.avatar.setMaximumSize(QtCore.QSize(35, 35))
					self.ui.avatar.setMinimumSize(QtCore.QSize(35, 35))
				else:
					pixmap.scaledToWidth(50)
					pixmap.scaledToHeight(50)
					self.ui.avatar.setMaximumSize(QtCore.QSize(50, 50))
					self.ui.avatar.setMinimumSize(QtCore.QSize(50, 50))
				self.ui.avatar.setPixmap(pixmap)
	
	def get_guid(self):
		return self.guid
		
	def link(self, link):
		print("link {}".format(link))
	
	def set_content(self, content):
		self.ui.html_content.document.setHtml(content)
	
	def get_content(self):
		if self.ui.content:
			return self.ui.html_content.document.toHtml()
		return None
