"""
This file is part of LionsTooth.

LionsTooth is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

LionsTooth is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with LionsTooth.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import pyqtSignal

from design.settings.ui_cache import Ui_cache

import os
import time # till now only need time so could do import time from time

class cache(QtWidgets.QGroupBox):
	fallback_cache_dir = './cache'
	one_day = 60*60*24
	default_config={
		'cache_dir'			: fallback_cache_dir,
		'clear_cache_days'	: 7,
		'clear_all_cache'	: False,
		'clear_on_start'	: True,
	}

	config_updated = pyqtSignal(str,object) # key, dict
	def __init__(self, config=None, parent=None):
		QtWidgets.QGroupBox.__init__(self, parent)
		self.ui = Ui_cache()
		self.ui.setupUi(self)

		if 'cache_dir' in config: self.config = config # TODO check if valid config
		else: self.config = self.default_config
		self.create_cache_dir() # try to create cache directory if it doesn't exists.

		self.files = []

		""" set values """
		self.ui.sb_clear_cache_days.setValue( int(self.config['clear_cache_days']) )
		
		checked = False
		if self.config['clear_all_cache'] in ['true', 'True', True]: checked = True# TODO this should be handled in config_handler
		self.ui.cb_clear_all_cache.setChecked( checked )
		
		checked = False
		if self.config['clear_on_start'] in ['true', 'True', True]: checked = True
		self.ui.cb_clear_on_start.setChecked( checked )

		if checked: self.clear_cache()
		else: self.update_values()

		""" Hooks """
		self.ui.sb_clear_cache_days.valueChanged.connect( self.set_clear_cache_days )
		self.ui.btn_clear_cache.clicked.connect( self.clear_cache ) # clear cache
		self.ui.btn_refresh.clicked.connect( self.update_values ) # refresh
		
		self.ui.cb_clear_on_start.stateChanged.connect( self.clear_on_start_toggled )

	def clear_on_start_toggled(self, value):
		if value: value = 'True'
		else: value = 'False'
		self.config['clear_on_start'] = value
		self.config_updated.emit('cache', self.config)

	def set_clear_cache_days(self, value):
		self.config['clear_cache_days'] = str(value)
		self.config_updated.emit('cache', self.config)
		self.update_values()

	def clear_cache(self):
		clear_all = False
		if self.ui.cb_clear_all_cache.checkState():
			clear_all = True
			self.ui.cb_clear_all_cache.setChecked(False)
		self.update_values(clear_all=clear_all)
		self.remove_files( self.files )
		self.update_values()

	def remove_files(self, files):
		for file_path in files:
			try:
				os.remove(file_path)
			except OSError:
				print("failed to remove: {}".format( file_path ))

	def update_values(self, clear_all=False):
		if clear_all:
			older_then = 0
		else:
			older_then = self.one_day * int(self.config['clear_cache_days'])
		self.files, cache_size, older_then_cache_size = self.get_cache_files(older_then=older_then)
		cache_size_mb = int(cache_size/(1000*1000))
		older_then_cache_size_mb = int(older_then_cache_size/(1000*1000))
		
		self.set_cache_size_label( cache_size_mb )
		self.set_cache_size_files_label( older_then_cache_size_mb )

	def set_cache_size_files_label(self, mb):
		self.ui.lbl_cache_size_files_value.setText("{} MB".format( mb ))

	def set_cache_size_label(self, mb):
		self.ui.lbl_cache_size_value.setText("{} MB".format( mb ))

	def create_cache_dir(self): # TODO set cache dir in config (also make sure its writeable then..)
		from os import path, makedirs
		if not path.exists(self.config['cache_dir']):
			makedirs(self.config['cache_dir'])

	def get_cache_files(self, older_then=0):
		""" This function will return a list with file path's, a int 
		refering to the total cache size in bytes and a int refering to the 
		size of files it returned.
		
		param cache_path: path to cache directory
		type cache_path: str
		
		param older_then: if this param is set it will only return the 
							path's of files that are older then x seconds.
							also the older_then_cache_size will probably 
							return something greater then 0 unless there 
							are none files that are older then x seconds.
		type older_then: int (seconds)
		
		"""
		now = int(time.time()) # timestamp
		cache_size = 0
		older_then_cache_size = 0
		files = []
		""" Only do directories in cache_path """
		for item in os.listdir( self.config['cache_dir'] ):
			path = os.path.join(self.config['cache_dir'], item)
			if os.path.isdir( path ):
				dir_size = 0
				for file in os.listdir(path):
					file_path = os.path.join(path, file)

					""" Only do files in sub directory, so not processing 
					subdirectories in this subdir, 'cause there should be 
					none. So recursive depth is 1. """
					if os.path.isfile( file_path ):
						last_modified = os.path.getmtime(file_path)
						file_size = os.path.getsize(file_path)
						""" If older_then is set the list with file path's
						it will return only will contain file path's of 
						files older then older_then. """
						if int(now - last_modified) > older_then:
							days_old = int((now - int(last_modified)) / (60*60*24))
							older_then_cache_size += file_size
							files.append( file_path )
						dir_size += file_size
				cache_size += dir_size
		return files, int(cache_size), int(older_then_cache_size)
