"""
This file is part of LionsTooth.

LionsTooth is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

LionsTooth is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with LionsTooth.  If not, see <http://www.gnu.org/licenses/>.
"""
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import pyqtSignal

class tags_box(QtWidgets.QFrame):
	tag_clicked = pyqtSignal(str) # tag_name
	def __init__(self, max_width=100, parent=None):
		QtWidgets.QFrame.__init__(self, parent)
		
		lo = QtWidgets.QHBoxLayout(self)
		self.max_width = max_width
		
		self.view_tags = {} # name: widget
		
		self.setLayout( lo )
	
	def set_max_width(self, max_width, change_layout=True):
		self.max_width = max_width
		for t_name in self.view_tags:
			self.view_tags[t_name].set_max_width( max_width )
		if change_layout: self.change_layout()
	
	def add_tag(self, name):
		if name not in self.view_tags:
			self.view_tags.update( { name: view_tag(name, max_width=self.max_width) } )
			self.view_tags[name].tag_clicked.connect( self.tag_clicked )

		self.layout().addWidget( self.view_tags[name], 0, QtCore.Qt.AlignLeft )
	
	def clear_layout(self):
		for name in self.view_tags:
			self.view_tags[name].deleteLater()
		self.view_tags = {}
		
	def change_layout(self):
		lo = self.layout()
		if type(lo) is QtWidgets.QHBoxLayout and self.size().width() >= self.max_width:
			# delete current layout
			QtCore.QObjectCleanupHandler().add(self.layout())
			
			# create new layout
			new_layout = QtWidgets.QVBoxLayout(self)
			
			# add widgets again
			for t_name in self.view_tags:
				self.add_tag( t_name )
			
			# set new layout
			self.setLayout( new_layout )

		elif type(lo) is QtWidgets.QVBoxLayout and self.size().width() < self.max_width:
			# delete current layout
			QtCore.QObjectCleanupHandler().add(self.layout())
			
			# create new layout
			new_layout = QtWidgets.QHBoxLayout(self)
			
			# add widgets again
			for t_name in self.view_tags:
				self.add_tag( t_name )
			
			# set new layout
			self.setLayout( new_layout )

class view_tag(QtWidgets.QLabel):
	tag_clicked = pyqtSignal(str) # tag name
	def __init__(self, name, max_width=100, parent=None):
		QtWidgets.QLabel.__init__(self, parent)
		self.name = name
		self.max_width = max_width
		
		sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
		self.setSizePolicy(sizePolicy)

		name = self.word_wrap_anywhere(name)
		self.set_name(name)

		self.default_css = ("color: #fff;\n"
		"background-color: #2c2c2c;\n"
		"border: 1px solid #000;\n"
		"border-radius: 3px;")
		
		self.default_hover_css = ("color: #fff;\n"
		"background-color: #1a1a1a;\n"
		"border: 1px solid #000;\n"
		"border-radius: 3px;")
		
		self.setStyleSheet(self.default_css)
		
		self.setOpenExternalLinks(False)
		self.linkActivated.connect( self.emit_tag_clicked )
		
		self.setCursor( QtCore.Qt.PointingHandCursor )
	
	def emit_tag_clicked(self):
		self.tag_clicked.emit( self.name )
	
	def set_name(self, name):
		css = '<style>a:link { color: #fff; }</style>'
		self.setText("{}<a href=\"tag://{}\"><b>#{}</b></a>".format(css, name, name))
	
	def set_max_width(self, max_width):
		self.max_width = max_width
		new_text = self.word_wrap_anywhere(self.name)
		self.set_name(new_text)
	
	def enterEvent(self, event):
		self.setStyleSheet(self.default_hover_css)
		
	def leaveEvent(self, event):
		self.setStyleSheet(self.default_css)
	
	def word_wrap_anywhere(self, text):
		font = self.font()
		font_metrics = QtGui.QFontMetrics(font)
		text_size = font_metrics.size(0, text)
		
		if text_size.width() > self.max_width:	
			new_text = ""
			cur_line = ""
			for char in text:
				if font_metrics.size(0, cur_line+char).width() >= self.max_width:
					cur_line += '<br>'
					new_text += cur_line
					cur_line = ""
				
				cur_line += char
			new_text += cur_line
			return new_text
		return text
