"""
This file is part of LionsTooth.

LionsTooth is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

LionsTooth is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with LionsTooth.  If not, see <http://www.gnu.org/licenses/>.
"""

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtGui import QPainter
from PyQt5.QtCore import pyqtSignal

from design.ui_side_menu import Ui_side_menu
from design.ui_stream_selector import Ui_stream_selector

class icon_button(QtWidgets.QFrame):
	clicked = pyqtSignal(bool)
	def __init__(self, text=None, icon_path=None, parent=None):
		QtWidgets.QFrame.__init__(self, parent)

		self.layout = QtWidgets.QHBoxLayout(self)
		self.layout.setContentsMargins(-1, 2, -1, 2)
		self.layout.setSpacing(0)

		self.text = QtWidgets.QPushButton()
		self.icon = QtWidgets.QPushButton()
		
		self.text.setFlat(True)
		self.icon.setFlat(True)
		
		self.text.setMinimumHeight(35)
		self.text.setMaximumHeight(35)
		self.icon.setMinimumSize(QtCore.QSize(35, 35))
		self.icon.setMaximumSize(QtCore.QSize(35, 35))
		
		self.text.setStyleSheet("text-align:left")
		
		self.layout.setContentsMargins(0, 0, 0, 0)
		self.layout.setSpacing(0);
		
		if text:
			self.set_text(text)
		
		if icon_path:
			self.set_icon(icon_path)

		self.layout.addWidget(self.text)
		self.layout.addWidget(self.icon, 0, QtCore.Qt.AlignRight|QtCore.Qt.AlignTop)
		
		self.checked = False
		
		self.text.clicked.connect( self.click )
		self.icon.clicked.connect( self.click )

	def set_icon(self, path):
		icon = QtGui.QIcon()
		icon.addPixmap(QtGui.QPixmap( path ), QtGui.QIcon.Normal, QtGui.QIcon.Off)
		self.icon.setIconSize(QtCore.QSize(24, 24))
		self.icon.setIcon(icon)

	def set_text(self, text):
		self.text.setText(text)
		
	def expand(self):
		self.text.show()
		
	def collapse(self):
		self.text.hide()
		
	def click(self):
		if self.checked:
			self.setChecked(False)
		else:
			self.setChecked(True)
		self.clicked.emit(self.checked)
		
	def isChecked(self):
		return self.checked
		
	def setChecked(self, state):
		if state:
			self.checked = True
			palette = self.text.palette().midlight().color()
			self.setStyleSheet("background-color: rgb({}, {}, {}, {});".format(palette.red(), palette.green(), palette.blue(), palette.alpha()))
		else:
			self.checked = False
			self.setStyleSheet("background-color: rgb(255, 255, 255, 0);")
		#print("icon_button.setChecked {} {}".format(state, self.checked)) # DEBUG

class delete_tag(QtWidgets.QPushButton):
	del_id = pyqtSignal(int) # int: id
	del_name = pyqtSignal(str) # int: tag name
	def __init__(self, id=None, name=None, parent=None):
		QtWidgets.QPushButton.__init__(self, parent)
		self.id = id
		self.name = name

		self.setMaximumSize(QtCore.QSize(20, 10))
		self.setFlat(True)
		self.setObjectName("btn_delete_tag")
		self.setText("X")

		if id:
			self.clicked.connect( self.emit_id )
		elif name:
			self.clicked.connect( self.emit_name )
	
	def emit_id(self):
		self.del_id.emit(self.id)
	
	def emit_name(self):
		self.del_name.emit(self.name)

class view_tag(QtWidgets.QPushButton):
	view_id = pyqtSignal(int, str) # int: id, str: name
	def __init__(self, id, name, parent=None):
		QtWidgets.QPushButton.__init__(self, parent)
		self.id = id
		self.name = name
		
		self.setFlat(True)
		self.setObjectName("btn_tag")
		self.setText("#{}".format(name))
		self.setCheckable(True)
		
		self.clicked.connect( self.emit_id )
	
	def emit_id(self):
		self.view_id.emit(self.id, self.name)		

class aspect_filter(QtWidgets.QCheckBox):
	def __init__(self, id, name, selected=False, parent=None):
		QtWidgets.QCheckBox.__init__(self, parent)

		self.name = name
		self.id = id

		self.setStyleSheet("background-color: rgba(255, 255, 255, 0);")
		self.setChecked( selected )
		self.setText( name )
		self.setTristate(False)

class stream_selector(QtWidgets.QWidget):
	state_changed = pyqtSignal(bool)
	selected = pyqtSignal(str)
	deselected = pyqtSignal(str)
	aspects_state_changed = pyqtSignal(bool)
	followed_tags_state_changed = pyqtSignal(bool)
	delete_tag = pyqtSignal(int) # tag id
	tag_clicked = pyqtSignal(str) # str: name
	request_follow_tag = pyqtSignal(str)
	filter_aspects = pyqtSignal()
	def __init__(self, parent=None):
		QtWidgets.QWidget.__init__(self, parent)

		self.ui = Ui_stream_selector()
		self.ui.setupUi(self)

		self.tags = {}
		self.viewed_tags = []
		self.aspects = {}

		self.selected_tag_id = None
		self.selected_aspects = []
		self.current_stream_selected_aspects = []

		self.close_aspects()
		self.close_tags()

		self.ui.recent_label.hide()

		# Stream button
		self.ui.btn_stream = icon_button(text="Stream", icon_path=":/icons/icons/steam_48.svg")
		self.ui.layout.insertWidget(0, self.ui.btn_stream)

		self.selected_button_name = None
		self.buttons = { 'stream': self.ui.btn_stream,
						'activity': self.ui.btn_activity,
						'aspects': self.ui.btn_aspects, 
						'mentions': self.ui.btn_mentions, 
						'public': self.ui.btn_public, 
						'followed_tags': self.ui.btn_tags,
						'load_tags': self.ui.btn_load_tags,
						}

		self.buttons['stream'].clicked.connect(lambda checked: self.button_state_changed('stream', checked) )
		self.buttons['activity'].clicked.connect(lambda checked: self.button_state_changed('activity', checked) )
		self.buttons['aspects'].clicked.connect(lambda checked: self.button_state_changed('aspects', checked) )
		self.buttons['mentions'].clicked.connect(lambda checked: self.button_state_changed('mentions', checked) )
		self.buttons['public'].clicked.connect(lambda checked: self.button_state_changed('public', checked) )
		self.buttons['followed_tags'].clicked.connect(lambda checked: self.button_state_changed('followed_tags', checked) )
		self.buttons['load_tags'].clicked.connect(lambda checked: self.button_state_changed('load_tags', checked) )

		self.ui.btn_add_tag.clicked.connect( self.add_new_tag )
		self.ui.btn_filter_aspects.clicked.connect( self.load_filtered )

	def deselect(self):
		for name, button in self.buttons.items():
			if name == self.selected_button_name:
				self.deselected.emit(name)
				self.buttons[name].setChecked( False )

	def close_all(self):
		self.close_aspects()
		self.close_tags()

	def button_state_changed(self, button_name, checked):
		for name, button in self.buttons.items():
			if button_name != name:
				if self.buttons[name].isChecked():
					self.buttons[name].setChecked( False ) # Deselect
					self.deselected.emit(button_name)
			else:# if button is presses once more while it's already checked
				if self.selected_button_name != name: # if different button, emit it
					self.selected_button_name = name
					if name != 'followed_tags':
						self.close_tags()
					if name != 'aspects':
						self.close_aspects()
				if not checked: # default behaviour will uncheck when pressed again
					self.buttons[name].setChecked( True )
				self.selected.emit(button_name)

	def open_aspects(self):
		self.ui.aspects_frame.show()
		self.aspects_state_changed.emit(True)

	def close_aspects(self):
		self.ui.aspects_frame.hide()
		self.aspects_state_changed.emit(False)

	def add_aspect(self, name, id, selected):
		if name not in self.aspects:
			self.aspects.update( { id: aspect_filter( id, name, selected, parent=self.ui.aspects_frame ) } )

			if selected:
				self.selected_aspects.append(id)
			self.ui.aspect_layout.addWidget( self.aspects[id] )

	def selected_aspect_ids(self):
		selected_aspect_ids = []
		for aspect_name in self.aspects:
			if self.aspects[aspect_name].checkState():
				selected_aspect_ids.append( self.aspects[aspect_name].id )
		return selected_aspect_ids

	def load_filtered(self):
		self.filter_aspects.emit()

	def del_aspect(self):
		pass
	
	def change_state(self, state):
		if state:
			self.ui.extra_frame.show()
		else:
			self.ui.extra_frame.hide()
	
	def close_tags(self):
		self.ui.tags_frame.hide()
		self.followed_tags_state_changed.emit(False)
		self.deselect_tag()

	def open_tags(self):
		self.ui.tags_frame.show()
		self.followed_tags_state_changed.emit(True)
		
	def tags_frame_state(self):
		return self.ui.tags_frame.isHidden()

	def add_new_tag(self, name=None):
		# TODO check if tag is valid..
		if not name:
			name = self.ui.edit_add_tag.text()
		self.request_follow_tag.emit( name )
		self.ui.edit_add_tag.setText("")

	def add_new_viewed_tag(self, name):
		if name not in self.viewed_tags:
			self.viewed_tags.append(name)
			self.add_viewed_tag(name)

		if self.viewed_tags:
			self.ui.recent_label.show()
		
	def add_viewed_tag(self, name):
		layout = QtWidgets.QHBoxLayout()
		layout.setObjectName("tag_layout_{}".format(name))
		btn_tag = view_tag( 0, name, parent=self.ui.tags_frame )
		btn_delete = delete_tag( name=name, parent=self.ui.tags_frame )
		
		btn_tag.view_id.connect(lambda: self.tag_clicked.emit( name ) )
		btn_delete.del_name.connect( self.del_viewed_tag )
		
		layout.addWidget( btn_tag )
		layout.addWidget( btn_delete )
		
		self.ui.recent_tags_layout.addLayout( layout )
		
		self.select_viewed_tag(name)

	def select_viewed_tag(self, name): # TODO subclass QHBoxLayout and put both buttons in it.
		child = self.findChild(QtWidgets.QHBoxLayout, "tag_layout_{}".format(name))
		if child:
			child = self.findChild(QtWidgets.QPushButton, "btn_tag")
			if child:
				child.setChecked(True)

	def del_viewed_tag(self, name):
		child = self.findChild(QtWidgets.QHBoxLayout, "tag_layout_{}".format(name))
		if child:
			child.deleteLater()
		
		if not self.viewed_tags:
			self.ui.recent_label.hide()

	def new_tag_followings(self, new_tag_followings):
		for i in range(len(new_tag_followings._tags)):
			if new_tag_followings[i].id not in self.tags:
				self.add_tag(new_tag_followings[i], index=i+1)# +1 to skip new tag input field and submit button

	def add_tag(self, tag=None, name=None, index=None): # TODO do something with the id and taggings_count
		if tag:
			name = tag.name
			id = tag.id
			taggings_count = tag.taggings_count
		else:
			id = None

		if id not in self.tags:
			self.tags.update( { id: { 
										'tag'				: tag,
										'layout'			: QtWidgets.QHBoxLayout(),
										'btn_tag'			: view_tag( id, name, self.ui.tags_frame ),
										'btn_delete_tag'	: delete_tag( id, self.ui.tags_frame ),
									}
			} )
			
			self.tags[id]['btn_delete_tag'].del_id.connect( self.del_tag )# connect delete button
			self.tags[id]['btn_tag'].view_id.connect( self.view_tag_id ) # connect view button
			
			self.tags[id]['layout'].addWidget( self.tags[id]['btn_tag'] )
			self.tags[id]['layout'].addWidget( self.tags[id]['btn_delete_tag'] )
			if index != None:
				self.ui.tags_layout.insertLayout(index, self.tags[id]['layout'] )
			else:
				self.ui.tags_layout.addLayout( self.tags[id]['layout'] )
		return id

	def deselect_tag(self):
		if self.selected_tag_id:
			if self.selected_tag_id in self.tags:
				self.tags[self.selected_tag_id]['btn_tag'].setChecked(False)
			self.selected_tag_id = None

	def view_tag_id(self, id, name):
		self.deselect_tag()
		self.tags[id]['btn_tag'].setChecked(True)
		self.selected_tag_id = id
		self.tag_clicked.emit(name)

	def del_tag(self, id):
		if self.tags[id]['tag']: # self.tags[id]['tag'] is non on newly created tags, those have no id and tag obj
			self.tags[id]['tag'].delete()# to pod
		if self.tags[id]['layout'] is not None:
			while self.tags[id]['layout'].count():
				item = self.tags[id]['layout'].takeAt(0)
				widget = item.widget()
				if widget is not None:
					widget.deleteLater()
		self.tags.pop(id)

class side_menu(QtWidgets.QWidget):
	state_changed = pyqtSignal(bool)
	selected = pyqtSignal(str) # button name as key in self.buttons
	deselected = pyqtSignal(str) # button name as key in self.buttons
	
	def __init__(self, parent=None):
		QtWidgets.QWidget.__init__(self, parent)
		self.background = QtGui.QColor(255, 255, 255)
		
		self.ui = Ui_side_menu()
		self.ui.setupUi(self)
		
		self.width = 300
		self.min_width = self.ui.item_layout.getContentsMargins()[0] + self.ui.item_layout.getContentsMargins()[2] + self.ui.btn_side_menu.size().width();

		self.move(0, 0)
		
		self.stream_selector = stream_selector()
		self.ui.item_layout.insertWidget(1, self.stream_selector)
		self.state_changed.connect( self.stream_selector.change_state )
		
		self.selected_button_name = None
		self.buttons = {} # TODO ordered dict
		self.buttons.update( { 'profile': icon_button(text="Profile", icon_path=":/icons/icons/profile.svg") } )
		self.buttons.update( { 'contacts': icon_button(text="Contacts", icon_path=":/icons/icons/contacts.svg") } )
		self.buttons.update( { 'settings': icon_button(text="Settings", icon_path=":/icons/icons/settings_48.svg") } )
		
		self.buttons['profile'].clicked.connect(lambda checked: self.button_state_changed('profile', checked) )
		self.buttons['settings'].clicked.connect(lambda checked: self.button_state_changed('settings', checked) )
		self.buttons['contacts'].clicked.connect(lambda checked: self.button_state_changed('contacts', checked) )
		
		count = 0
		for button_name in self.buttons:
			self.ui.item_layout.insertWidget(2+count, self.buttons[button_name])
			count +=1

		# Collapse and Expand side menu
		self.ui.btn_side_menu.clicked.connect(self.toggle_menu)

		# If button from stream selector gets selected then deselect:
		# settings button
		# profile button
		self.stream_selector.selected.connect( self.stream_selector_button_clicked )

	def stream_selector_button_clicked(self):
		for button_name in self.buttons:
			self.buttons[button_name].setChecked(False)

	def button_state_changed(self, button_name, checked):
		self.stream_selector.deselect() # deselect stream_selector item
		for name, button in self.buttons.items():
			if button_name != name:
				if self.buttons[name].isChecked():
					self.buttons[name].setChecked( False ) # Deselect
					self.deselected.emit(button_name)
			else:# if button is presses once more while it's already checked
				if self.selected_button_name != name: # if different button, emit it
					self.selected_button_name = name
				if not checked: # default behaviour will uncheck when pressed again
					self.buttons[name].setChecked( True )
				self.selected.emit(button_name)

	def deselect(self):
		self.stream_selector.deselect()
		for name, button in self.buttons.items():
			button.setChecked(False)

	def adjust_height(self, height):
		self.resize(self.size().width(), height)

	def paintEvent(self, event):
		size = self.size()
		painter = QPainter()
		painter.begin(self)
		painter.setRenderHint(QPainter.Antialiasing)
		painter.setBrush(self.background)
		
		painter.drawRect(0, 0, size.width(), size.height())
		painter.end()
		
	def resizeEvent(self, event):
		print("resize side_menu")
		
	def toggle_menu(self):
		size = self.size()
		if size.width() != self.min_width: # close
			self.resize(self.min_width, size.height())
			self.stream_selector.ui.btn_stream.collapse()
			for name, button in self.buttons.items(): # hide text
				button.collapse()
			self.state_changed.emit(False)
		else: # open
			self.resize(self.width, self.height())
			self.stream_selector.ui.btn_stream.expand()
			for name, button in self.buttons.items(): # show text
				button.expand()
			self.state_changed.emit(True)
