"""
This file is part of LionsTooth.

LionsTooth is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

LionsTooth is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with LionsTooth.  If not, see <http://www.gnu.org/licenses/>.
"""

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtGui import QPainter
from PyQt5.QtCore import pyqtSignal

from design.ui_stream import Ui_stream
from widgets.post_widget import post_widget
from widgets.shared.loading_icon import loading_icon

class stream(QtWidgets.QWidget):
	updated = pyqtSignal(bool)
	author_clicked = pyqtSignal(str) # str: author guid
	anchor_clicked = pyqtSignal(str) # str: link
	permalink_clicked = pyqtSignal(str) # str: post guid
	new_stream = pyqtSignal()
	scroll_value_changed = pyqtSignal(str, int) # self.name, value
	own_avatar_updated = pyqtSignal(str)# avatar path (local)
	res_set = pyqtSignal()
	search_set = pyqtSignal()
	post_widget_inserted = pyqtSignal(str) # post guid
	post_widgets_inserted = pyqtSignal()
	max_width_set = pyqtSignal(int) # max_width

	def __init__(self, parent=None):
			QtWidgets.QWidget.__init__(self, parent)
			self.ui = Ui_stream()
			self.ui.setupUi(self)

			self.res = None
			self.search = None
			self.stream = None
			self.name = None

			self.ui.stream_layout.post_widgets = {}
			self.id_to_guid = {} # used to get guid from id
			self.ui.loading_icon_top = None
			self.ui.loading_icon_bottom = None
			self.loading_icon_count = [False,False]# top, bottom
			self.post_offset = 0
			self.max_width = 200
			self.own_guid = None
			self.own_avatar = None

			""" When posts are inserted, it will first empty and then 
			write new guids to this (so when more posts are requested it
			can be used to detect when there are now posts left.)"""
			self.new_guids = []
			self.more_fail_count = 0
			self.more_fail_count_max = 2

			self.go_to_anchor_que = []

			self.ui.scrollArea.verticalScrollBar().valueChanged.connect( self.scroll_changed )
			self.ui.scrollArea.verticalScrollBar().valueChanged.connect( lambda value: self.scroll_value_changed.emit(self.name, value) )

	def set_own_avatar(self, path):
		self.own_avatar = path
		self.own_avatar_updated.emit(path)

	def update_id_to_guid(self):
		if self.ui.stream_layout.post_widgets:
			id_to_guid = {}
			for guid, widget in self.ui.stream_layout.post_widgets:
				id_to_guid.update( { widget.id: widget.guid } )
			self.id_to_guid = id_to_guid

	def process_anchor_que(self):
		for params in self.go_to_anchor_que:
			self.go_to_anchor(post_guid=params[0], comment_id=params[1])
		self.go_to_anchor_que = []

	def go_to_anchor(self, post_guid=None, comment_id=None):
		""" This function will set the vertical scrollbar so that the 
		post with the provided guid is set to top ( or comment of that 
		post if comment_id is also provided )

		TODO
		Here check if there is a vertical scrollbar
		if not verticascroll:
			return
		else:
		"""
		# Check if we got the guid or id if it's the id then get the guid
		if type(post_guid) is int: # it's id
			self.update_id_to_guid()
			post_guid = self.id_to_guid[post_guid]

		# If post guid not in stream_layout's post_widgets then return
		if post_guid not in self.ui.stream_layout.post_widgets: return
		p = self.ui.stream_layout.post_widgets[post_guid]

		# Set scroll_value (y) to position of post_widget in scrollArea
		scroll_value = p.pos().y()

		if comment_id and p.comments._comments:
			p.comments.adjustSize() # <-- needed, else we will get wrong values
			# Add comments pos in post_widget to scroll_value (y)
			scroll_value += p.comments.pos().y()

			# Check if the comment_id param is a str, then we expect a str
			if type(comment_id) is str: # guid
				p.comments.update_guid_to_id()
				comment_id = p.comments.get_id( comment_id )

			# Show all comments ('cause we currently don't check if comment with comment_id is in the layout)
			if p.comments.is_collapsed():
				p.comments.show_more()

			# Get current comments
			current_comments = p.comments.current_comments()

			# Add comment pos in comments to scroll_value (y)
			if comment_id:
				scroll_value += current_comments[comment_id].pos().y()
				current_comments[comment_id].select() # set stylesheet

		# If verticalScrollBar has no maximum, set it to scroll_value
		if not self.ui.scrollArea.verticalScrollBar().maximum():
			self.ui.scrollArea.verticalScrollBar().setMaximum( scroll_value )

		# Set scroll position to scroll_value
		self.ui.scrollArea.verticalScrollBar().setValue( scroll_value )

	def clear_layout(self):
		for i in range(0,self.ui.stream_layout.count()):# > self.post_offset:
			if i < self.post_offset:
				continue
			item = self.ui.stream_layout.takeAt(self.post_offset)
			if not item:
				continue
			w = item.widget()
			if w:
				if type(w) is post_widget:
					w.author_clicked.disconnect( self.author_clicked )
					w.anchor_clicked.disconnect( self.anchor_clicked )
					w.permalink_clicked.disconnect( self.permalink_clicked )
					w.deleteLater()
		self.ui.stream_layout.post_widgets = {}

	def reset(self):
		self.clear_layout()
		self.res = None
		self.stream = None
		self.own_guid = None

	def set_res(self, res):
		self.res = res
		self.res_set.emit()

	def set_search(self, search):
		self.search = search
		self.search_set.emit()

	def set_stream(self, stream):
		self.stream = stream
		self.new_stream.emit()

	def set_own_guid(self, own_guid):
		self.own_guid = own_guid

	def set_max_width(self, max_width):
		self.max_width = max_width
		self.max_width_set.emit(max_width)

	def current_posts(self):
		return self.ui.stream_layout.post_widgets 

	def insert_post_widget(self, post, index_offset=0, show_all_comments=False):
		uid = post.guid # unique id. wheter it's guid or id
		if not post.guid:
			uid = post.id
		if (post.guid in self.ui.stream_layout.post_widgets or 
			post.id in self.ui.stream_layout.post_widgets or # newly posted post returns only id, if we don't check this we get duplicates on screen.
			not uid):
			return None

		p = post_widget(
			self.res, 
			post=post, 
			page_name=self.name, 
			thread_pool=self.thread_pool, 
			own_guid=self.own_guid, 
			own_avatar=self.own_avatar, 
			max_width=self.max_width, 
			parent=self.ui.main,
			show_all_comments=show_all_comments)

		p.author_clicked.connect( self.author_clicked ) # set author name clicked event
		p.anchor_clicked.connect( self.anchor_clicked ) # when clicked on a hyperlink
		p.permalink_clicked.connect( self.permalink_clicked )

		self.ui.stream_layout.post_widgets.update( { uid: p } )

		insert_index = index_offset + self.post_offset
		self.ui.stream_layout.insertWidget( insert_index, p )

		self.ui.stream_layout.post_widgets[uid].set_max_width( self.max_width )

		# Emit signal that we inserted post with guid.
		self.post_widget_inserted.emit(uid)

		return uid

	def insert_post_widgets(self):
		stream_copy = self.stream.copy()
		current_posts = self.current_posts()

		added_uids = [] # newly added guids
		new_uids = [] # use this to check if post is removed from stream (this contains all guids it received, old and new ones)
		for i in range(len(stream_copy)):
			post = stream_copy[i]

			uid = post.guid 
			if not post.guid:
				uid = post.id
			new_uids.append(uid)

			added_uid = self.insert_post_widget(post, i)

			if added_uid: # insert_post_widget will return None if it's already there
				added_uids.append( added_uid )

		self.new_guids = added_uids

		""" check if post is removed """
		for post_id in current_posts:
			if post_id not in new_uids:
				current_posts[post_id].deleteLater() # remove it

		# Emit signal that we are done.
		self.post_widgets_inserted.emit()

	def scroll_changed(self, value):
		for uid in self.ui.stream_layout.post_widgets:
			w = self.ui.stream_layout.post_widgets[uid]
			if not w.visibleRegion().boundingRect():
					if w.is_reshare:
						if w.reshare_post:
							if w.reshare_post.has_markdown_content:
								w.reshare_post.ui.markdown_viewer.stop_animations()
							if w.reshare_post.has_photo_browser and w.reshare_post.photo_browser.ui.photo.is_animation():
								w.reshare_post.photo_browser.ui.photo.stop_animation()
					else:
						if w.has_markdown_content:
							w.ui.markdown_viewer.stop_animations()
						if w.has_photo_browser and w.photo_browser.ui.photo.is_animation():
								w.photo_browser.ui.photo.stop_animation()				

	def start_top_loading(self):
		if not self.loading_icon_count[0]:
			print("__START_TOP_LOADING__")
			self.loading_icon_count[0] = True
			self.ui.loading_icon_top = loading_icon(self)
			self.ui.stream_layout.insertWidget(self.post_offset, self.ui.loading_icon_top, 0, QtCore.Qt.AlignCenter|QtCore.Qt.AlignTop)
			self.ui.loading_icon_top.start()

	def stop_top_loading(self):
		if self.loading_icon_count[0]:
			print("__STOP_TOP_LOADING__")
			try:
				self.loading_icon_count[0] = False
				self.ui.loading_icon_top.deleteLater()
				self.ui.loading_icon_top = None
			except RuntimeError:
				print("DEBUG RuntimeError stop_top_loading")

	def start_bottom_loading(self):
		if not self.loading_icon_count[1]:
			print("__START_BOTTOM_LOADING__")
			self.loading_icon_count[1] = True
			self.ui.loading_icon_bottom = loading_icon(self)
			self.ui.stream_layout.addWidget(self.ui.loading_icon_bottom, 0, QtCore.Qt.AlignCenter|QtCore.Qt.AlignBottom)
			self.ui.loading_icon_bottom.start()

	def stop_bottom_loading(self):
		if self.loading_icon_count[1]:
			print("__STOP_BOTTOM_LOADING__")
			try:
				self.loading_icon_count[1] = False
				self.ui.loading_icon_bottom.deleteLater()
				self.ui.loading_icon_bottom = None
			except RuntimeError:
				print("DEBUG RuntimeError stop_bottom_loading")
